#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""

from collections import abc
from collections.abc import Iterator, Sequence
from enum import Enum
from numbers import Number
from typing import Literal, Optional, Union

import numpy as np
from numpy.typing import NDArray

from pyxel.evaluator import eval_range


class ParameterType(Enum):
    """TBW."""

    # one-dimensional, can be a dataset coordinate
    Simple = "simple"
    # multi-dimensional or loaded from a file in parallel mode, cannot be a dataset coordinate
    Multi = "multi"


# TODO: Add unit tests. See #336
class ParameterValues:
    """Contains keys and values of parameters in a parametric step."""

    def __init__(
        self,
        key: str,
        values: Union[
            Literal["_"],
            Sequence[Literal["_"]],
            Sequence[Number],
            Sequence[str],
        ],
        boundaries: Union[
            tuple[float, float],
            Sequence[tuple[float, float]],
            None,
        ] = None,
        enabled: bool = True,
        logarithmic: bool = False,
    ):
        # TODO: maybe use numpy to check multi-dimensional input lists
        # Check YAML input (not real values yet) and define parameter type
        if values == "_":
            self.type: ParameterType = ParameterType("multi")
        elif isinstance(values, str) and "numpy" in values:
            self.type = ParameterType("simple")
        elif isinstance(values, abc.Sequence) and any(
            [(el == "_" or isinstance(el, abc.Sequence)) for el in values]
        ):
            self.type = ParameterType("multi")
        elif isinstance(values, abc.Sequence) and all(
            [isinstance(el, (Number, str)) for el in values]
        ):
            self.type = ParameterType("simple")
        else:
            raise ValueError("Parameter values cannot be initiated with those values.")

        if "sampling.start_time" in key:
            key = "detector.sampling_properties.start_time"

        # unique identifier to the step. example: 'detector.geometry.row'
        self._key: str = key
        self._values: Union[
            Literal["_"], Sequence[Literal["_"]], Sequence[Number], Sequence[str]
        ] = values

        # short  name identifier: 'row'
        self._short_name = key.split(".")[-1]

        self._enabled: bool = enabled
        self._logarithmic: bool = logarithmic

        if boundaries is None:
            boundaries_array: Optional[NDArray[np.float_]] = None
        else:
            boundaries_array = np.array(boundaries, dtype=np.float_)
            if boundaries_array.ndim == 1:
                if boundaries_array.shape != (2,):
                    raise ValueError(
                        f"Expecting only two values for the boundaries. Got: {boundaries}."
                    )
            elif boundaries_array.ndim == 2:
                if boundaries_array.shape != (len(values), 2):
                    raise ValueError(
                        f"Expecting a 2x{len(values)} values for the boundaries. Got: {boundaries}."
                    )
            else:
                raise ValueError(f"Wrong format of boundaries. Got {boundaries}.")

        self._boundaries: Optional[NDArray[np.float_]] = boundaries_array

        self._current: Optional[Union[Literal["_"], Number, str]] = None

    def __repr__(self) -> str:
        cls_name: str = self.__class__.__name__

        return f"{cls_name}<key={self.key!r}, values={self.values!r}, enabled={self.enabled!r}>"

    def __len__(self) -> int:
        try:
            values: Sequence = eval_range(self.values)
        except Exception as exc:
            raise NameError(
                f"Wrong value: {self.values!r} for key: {self.key!r}. Error: {exc}"
            ) from exc

        return len(values)

    # TODO: Is method '__contains__' needed ? If yes then this class will act as a `Collections.abc.Sequence`
    def __iter__(self) -> Iterator[Number]:
        values = eval_range(self.values)
        yield from values

    @property
    def key(self) -> str:
        """TBW."""
        return self._key

    @property
    def short_name(self) -> str:
        """TBW."""
        return self._short_name

    @property
    def values(
        self,
    ) -> Union[Literal["_"], Sequence[Literal["_"]], Sequence[Number], Sequence[str]]:
        """TBW."""
        return self._values

    @property
    def enabled(self) -> bool:
        """TBW."""
        return self._enabled

    @property
    def current(self) -> Optional[Union[Literal["_"], Number, str]]:
        """TBW."""
        return self._current

    @current.setter
    def current(self, value: Union[Literal["_"], str, Number]) -> None:
        """TBW."""
        self._current = value

    @property
    def logarithmic(self) -> bool:
        """TBW."""
        return self._logarithmic

    @property
    def boundaries(self) -> Optional[NDArray[np.float_]]:
        """TBW."""
        return self._boundaries



#   Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#  #
#   This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#   is part of this Pyxel package. No part of the package, including
#   this file, may be copied, modified, propagated, or distributed except according to
#   the terms contained in the file ‘LICENCE.txt’.

"""Remove cosmic rays from an astronomical image using the LA Cosmic algorithm.

References,

* https://esa.gitlab.io/pyxel/doc/stable/references/model_groups/data_processing_models.html
* http://www.astro.yale.edu/dokkum/lacosmic/
* https://lacosmic.readthedocs.io/en/stable/
* https://www.astropy.org/ccd-reduction-and-photometry-guide/v/dev/notebooks/08-03-Cosmic-ray-removal.html
"""
import numpy as np
import xarray as xr
from datatree import DataTree

from pyxel.detectors import Detector


def compute_cosmic_rays(
    data_2d: np.ndarray,
    contrast: float = 1.0,
    cr_threshold: float = 50.0,
    neighbor_threshold: float = 50.0,
    effective_gain: float = 1.0,
    readnoise: float = 0.0,
) -> tuple[np.ndarray, np.ndarray]:
    """Compute the image with cosmics removed, including the cosmic rays mask images.

    Refer to lacosmic documentation for parameter descriptions.
    https://lacosmic.readthedocs.io/en/stable/api/lacosmic.lacosmic.html#lacosmic.lacosmic

    Parameters
    ----------
    data_2d : np.ndarray
        2D image array
    contrast : float
        Contrast threshold between the Laplacian image and the fine-structure image.
    cr_threshold : float
        The Laplacian signal-to-noise ratio threshold for cosmic-ray detection.
    neighbor_threshold : float
        The Laplacian signal-to-noise ratio threshold for detection of cosmic rays in
        pixels neighboring the initially-identified cosmic rays.
    effective_gain : float
        Ratio of counts (e.g., electrons or photons) to the units of data.
    readnoise : float
        The read noise (in electrons) in the input data.
    """
    import lacosmic

    cleaned_image, cr_mask = lacosmic.lacosmic(
        data_2d,
        contrast=contrast,
        cr_threshold=cr_threshold,
        neighbor_threshold=neighbor_threshold,
        effective_gain=effective_gain,
        readnoise=readnoise,
    )

    return cleaned_image, cr_mask


def remove_cosmic_rays(
    detector: Detector,
    contrast: float = 1.0,
    cr_threshold: float = 50.0,
    neighbor_threshold: float = 50.0,
    effective_gain: float = 1.0,
    readnoise: float = 0.0,
) -> None:
    """Extract the roi data converts it to xarray dataset and saves the information to the final result.

    Refer to lacosmic documentation for parameter descriptions.
    https://lacosmic.readthedocs.io/en/stable/api/lacosmic.lacosmic.html#lacosmic.lacosmic

    Parameters
    ----------
    detector : Detector
        Pyxel Detector object.
    contrast : float
        Contrast threshold between the Laplacian image and the fine-structure image.
    cr_threshold : float
        The Laplacian signal-to-noise ratio threshold for cosmic-ray detection.
    neighbor_threshold : float
        The Laplacian signal-to-noise ratio threshold for detection of cosmic rays in
        pixels neighboring the initially-identified cosmic rays.
    effective_gain : float
        Ratio of counts (e.g., electrons or photons) to the units of data.
    readnoise : float
        The read noise (in electrons) in the input data.
    """
    cleaned_image, cr_mask = compute_cosmic_rays(
        data_2d=detector.pixel.array,
        contrast=contrast,
        cr_threshold=cr_threshold,
        neighbor_threshold=neighbor_threshold,
        effective_gain=effective_gain,
        readnoise=readnoise,
    )
    detector.pixel.array = cleaned_image

    # Get current absolute time
    absolute_time = xr.DataArray(
        [detector.absolute_time],
        dims=["time"],
        attrs={"units": "s"},
    )
    dataset = xr.Dataset()
    dataset["cosmic_ray_mask"] = xr.DataArray(cr_mask, dims=["y", "x"])
    dataset["cosmic_ray_clean"] = detector.pixel.to_xarray()
    dataset = dataset.expand_dims(dim="time").assign_coords(time=absolute_time)

    key = "lacosmic"
    key_partial = "lacosmic_partial"

    try:
        _ = detector.data[key_partial]
    except KeyError:
        has_key_partial = False
    else:
        has_key_partial = True

    if not has_key_partial:
        data_tree: DataTree = DataTree(dataset)
    else:
        # Concatenate data
        previous_datatree = detector.data[key_partial]
        data_tree = previous_datatree.combine_first(dataset)  # type: ignore

    if detector.pipeline_count == (detector.num_steps - 1):
        detector.data[key] = data_tree
    else:
        detector.data[key_partial] = data_tree

    # This is the last step and there is at least two steps
    if detector.num_steps > 1 and (detector.pipeline_count == (detector.num_steps - 1)):
        detector.data[key_partial].orphan()


# TODO: document the parameters and reference lacosmic
# TODO: investigate if astroscrappy can be used. It's provided with anaconda.
# TODO: add a switch to choose between astroscrappy and lacosmic



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import pickle
from pathlib import Path

import cloudpickle
import pytest

import pyxel
from pyxel.pipelines import DetectionPipeline


@pytest.mark.deprecated
@pytest.fixture
def pipeline_single_deprecated() -> DetectionPipeline:
    filename_single = Path("tests/data/deprecated_yaml.yaml")
    assert filename_single.exists()

    cfg = pyxel.load(filename_single)
    pipeline: DetectionPipeline = cfg.pipeline

    return pipeline


@pytest.fixture
def pipeline_single() -> DetectionPipeline:
    filename_single = Path("tests/data/yaml.yaml")
    assert filename_single.exists()

    cfg = pyxel.load(filename_single)
    pipeline: DetectionPipeline = cfg.pipeline

    return pipeline


@pytest.mark.deprecated
@pytest.mark.parametrize("pickle_method", [pickle, cloudpickle])
def test_serialization_deprecated(
    pipeline_single_deprecated: DetectionPipeline, pickle_method
):
    """Test serialization/deserialization."""
    data = pickle_method.dumps(pipeline_single_deprecated)
    assert isinstance(data, bytes)

    obj = pickle_method.loads(data)
    assert isinstance(obj, DetectionPipeline)
    assert obj is not pipeline_single_deprecated


@pytest.mark.parametrize("pickle_method", [pickle, cloudpickle])
def test_serialization(pipeline_single: DetectionPipeline, pickle_method):
    """Test serialization/deserialization."""
    data = pickle_method.dumps(pipeline_single)
    assert isinstance(data, bytes)

    obj = pickle_method.loads(data)
    assert isinstance(obj, DetectionPipeline)
    assert obj is not pipeline_single



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""

import warnings
from collections.abc import Iterable, Sequence
from typing import Optional

from pyxel.pipelines import ModelFunction, ModelGroup


class DetectionPipeline:
    """TBW."""

    # Define the order of steps in the pipeline.
    MODEL_GROUPS: tuple[str, ...] = (
        "scene_generation",
        "photon_collection",
        "photon_generation",  # Deprecated. It will be removed in version 2.0
        "optics",  # Deprecated. It will be removed in version 2.0
        "phasing",
        "charge_generation",
        "charge_collection",
        "charge_transfer",
        "charge_measurement",
        "signal_transfer",
        "readout_electronics",
        "data_processing",
    )

    # TODO: develop a ModelGroupList class ? See #333
    def __init__(
        self,  # TODO: Too many instance attributes
        scene_generation: Optional[Sequence[ModelFunction]] = None,
        photon_collection: Optional[Sequence[ModelFunction]] = None,
        photon_generation: Optional[Sequence[ModelFunction]] = None,  # Deprecated
        optics: Optional[Sequence[ModelFunction]] = None,  # Deprecated
        phasing: Optional[Sequence[ModelFunction]] = None,
        charge_generation: Optional[Sequence[ModelFunction]] = None,
        charge_collection: Optional[Sequence[ModelFunction]] = None,
        charge_transfer: Optional[Sequence[ModelFunction]] = None,
        charge_measurement: Optional[Sequence[ModelFunction]] = None,
        signal_transfer: Optional[Sequence[ModelFunction]] = None,
        readout_electronics: Optional[Sequence[ModelFunction]] = None,
        data_processing: Optional[Sequence[ModelFunction]] = None,
        doc: Optional[str] = None,
    ):
        self.doc = doc

        self._scene_generation: Optional[ModelGroup] = (
            ModelGroup(scene_generation, name="scene_generation")
            if scene_generation
            else None
        )

        self._photon_collection: Optional[ModelGroup] = (
            ModelGroup(photon_collection, name="photon_collection")
            if photon_collection
            else None
        )

        # Deprecated
        self._photon_generation: Optional[ModelGroup] = (
            ModelGroup(photon_generation, name="photon_generation")
            if photon_generation
            else None
        )

        # Deprecated
        self._optics: Optional[ModelGroup] = (
            ModelGroup(optics, name="optics") if optics else None
        )

        self._phasing: Optional[ModelGroup] = (
            ModelGroup(phasing, name="phasing") if phasing else None
        )  # MKID-array

        self._charge_generation: Optional[ModelGroup] = (
            ModelGroup(charge_generation, name="charge_generation")
            if charge_generation
            else None
        )

        self._charge_collection: Optional[ModelGroup] = (
            ModelGroup(charge_collection, name="charge_collection")
            if charge_collection
            else None
        )

        self._charge_measurement: Optional[ModelGroup] = (
            ModelGroup(charge_measurement, name="charge_measurement")
            if charge_measurement
            else None
        )

        self._readout_electronics: Optional[ModelGroup] = (
            ModelGroup(readout_electronics, name="readout_electronics")
            if readout_electronics
            else None
        )

        self._charge_transfer: Optional[ModelGroup] = (
            ModelGroup(charge_transfer, name="charge_transfer")
            if charge_transfer
            else None
        )  # CCD

        self._signal_transfer: Optional[ModelGroup] = (
            ModelGroup(signal_transfer, name="signal_transfer")
            if signal_transfer
            else None
        )  # CMOS

        self._data_processing: Optional[ModelGroup] = (
            ModelGroup(data_processing, name="data_processing")
            if data_processing
            else None
        )

    def __repr__(self) -> str:
        cls_name: str = self.__class__.__name__
        return f"{cls_name}<doc={self.doc!r}>"

    def __iter__(self) -> Iterable[ModelFunction]:
        for model in self.MODEL_GROUPS:
            models_grp: Optional[ModelGroup] = getattr(self, model)
            if models_grp:
                yield from models_grp

    @property
    def scene_generation(self) -> Optional[ModelGroup]:
        """Get group 'scene generation'."""
        return self._scene_generation

    @property
    def photon_collection(self) -> Optional[ModelGroup]:
        """Get group 'photon collection'."""
        return self._photon_collection

    @property
    def photon_generation(self) -> Optional[ModelGroup]:
        """Get group 'photon generation'."""
        warnings.warn(
            "Group 'photon_generation' is deprecated "
            "and will be removed in version 2.0. "
            "Use group 'photon_collection' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._photon_generation

    @property
    def optics(self) -> Optional[ModelGroup]:
        """Get group 'optics'."""
        warnings.warn(
            "Group 'optics' is deprecated "
            "and will be removed in version 2.0. "
            "Use group 'photon_collection' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._optics

    @property
    def phasing(self) -> Optional[ModelGroup]:
        """Get group 'phasing'."""
        return self._phasing

    @property
    def charge_generation(self) -> Optional[ModelGroup]:
        """Get group 'charge generation'."""
        return self._charge_generation

    @property
    def charge_collection(self) -> Optional[ModelGroup]:
        """Get group 'charge collection'."""
        return self._charge_collection

    @property
    def charge_transfer(self) -> Optional[ModelGroup]:
        """Get group 'charge transfer'."""
        return self._charge_transfer

    @property
    def charge_measurement(self) -> Optional[ModelGroup]:
        """Get group 'charge measurement'."""
        return self._charge_measurement

    @property
    def signal_transfer(self) -> Optional[ModelGroup]:
        """Get group 'signal transfer'."""
        return self._signal_transfer

    @property
    def readout_electronics(self) -> Optional[ModelGroup]:
        """Get group 'readout electronics'."""
        return self._readout_electronics

    @property
    def data_processing(self) -> Optional[ModelGroup]:
        """Get group 'data processing'."""
        return self._data_processing

    @property
    def model_group_names(self) -> tuple[str, ...]:
        """TBW."""
        return self.MODEL_GROUPS

    # TODO: Is this method used ?
    def get_model(self, name: str) -> ModelFunction:
        """Return a ModelFunction object for the specified model name.

        Parameters
        ----------
        name: str
            Name of the model.

        Returns
        -------
        ModelFunction

        Raises
        ------
        AttributeError
            If model with the specified name is not found.
        """
        group_name: str
        for group_name in self.model_group_names:
            model_group: ModelGroup = getattr(self, group_name)
            if model_group:
                model: ModelFunction
                for model in model_group.models:
                    if name == model.name:
                        return model
        raise AttributeError("Model has not been found.")



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Tests for load psf model."""
from pathlib import Path

import numpy as np
import pytest

from pyxel.detectors import (
    CCD,
    CCDGeometry,
    Characteristics,
    Detector,
    Environment,
    ReadoutProperties,
)
from pyxel.models.photon_collection import load_psf


@pytest.fixture
def ccd_10x10() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )
    detector._readout_properties = ReadoutProperties(times=[1.0])
    return detector


@pytest.fixture
def psf_10x10() -> np.ndarray:
    """Create a gaussian psf of 10x10."""
    shape = (10, 10)
    ny, nx = (el / 2 for el in shape)
    ay, ax = (np.arange(-el / 2.0 + 0.5, el / 2.0 + 0.5) for el in shape)
    xx, yy = np.meshgrid(ax, ay, indexing="xy")
    r = ((xx / nx) ** 2 + (yy / ny) ** 2) ** 0.5
    out = np.ones(shape)
    sigma = 0.5
    out[...] = np.exp(-(r**2) / (2 * sigma**2)) / (np.sqrt(2 * np.pi) * sigma)
    return out


@pytest.fixture
def psf_10x10_npy_filename(psf_10x10: np.ndarray, tmp_path: Path) -> Path:
    """Create a filename with a psf 10x10."""
    filename = tmp_path / "psf.npy"
    np.save(filename, psf_10x10)
    return filename


def test_load_psf(ccd_10x10: CCD, psf_10x10_npy_filename: Path) -> None:
    """Test function 'load_psf'."""
    detector: Detector = ccd_10x10

    # Run model
    load_psf(
        detector=detector,
        filename=psf_10x10_npy_filename,
    )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


"""Simple ADC model functions."""

from typing import Literal, Optional

import numpy as np
from numpy.typing import DTypeLike

from pyxel.detectors import Detector


def apply_simple_adc(
    signal: np.ndarray,
    bit_resolution: int,
    voltage_min: float,
    voltage_max: float,
    dtype: DTypeLike,
) -> np.ndarray:
    """Apply a simple Analog-to-Digital Converted (ADC) digitization.

    This functions simulates the behaviour of an ADC by quantizing a continuous signal
    to a discrete digital representation based on the provided bit resolution and voltage range.

    Parameters
    ----------
    signal : array
        Continuous input signal to be quantized.
    bit_resolution : int
        ADC bit resolution used for quantization.
    voltage_min : float
        The minimum voltage level in the input signal's range.
    voltage_max : float
        The maximum voltage level in the input signal's range.
    dtype : DTypeLike
        Desired data type of the output signal.

    Returns
    -------
    array
        The quantized digital representation.

    Examples
    --------
    >>> apply_simple_adc(
    ...     signal=np.array([-0.1, 0.0, 3.0, 6.0, 6.1]),
    ...     bit_resolution=8,
    ...     voltage_min=0.0,
    ...     voltage_max=6.0,
    ...     dtype=np.uint8,
    ... )
    array([0, 0, 127, 255, 255], dtype=np.uint8

    Notes
    -----
    This function performs the following steps:
    1. Clips the input signal to the specified voltage range [voltage_min, voltage_max].
    1. Normalizes the clipped signal to the range [0, 2^bit_resolution - 1].
    1. Rounds the normalized values to the nearest integer using truncation.
    1. Converts the resulting array to the specified data type (dtype).
    """
    output = (
        (np.clip(signal, a_min=voltage_min, a_max=voltage_max) - voltage_min)
        * (2**bit_resolution - 1)
        / (voltage_max - voltage_min)
    )

    return np.trunc(output).astype(dtype)


def _get_dtype(bit_resolution: int) -> np.dtype:
    """Get NumPy data type based on a given bit resolution.

    Parameters
    ----------
    bit_resolution : int
        Number of bits representing the data.

    Returns
    -------
    np.dtype
        Numpy data type corresponding to the provided bit resolution.

    Raises
    ------
    ValueError
        Raised if the bit resolution does not fall within the supported range [1, 64]

    Examples
    --------
    >>> _get_dtype(8)
    dtype('uint8')
    >>> _get_dtype(12)
    dtype('uint16')
    """
    if 1 <= bit_resolution <= 8:
        return np.dtype(np.uint8)
    elif 9 <= bit_resolution <= 16:
        return np.dtype(np.uint16)
    elif 17 <= bit_resolution <= 32:
        return np.dtype(np.uint32)
    elif 33 <= bit_resolution <= 64:
        return np.dtype(np.uint64)
    else:
        raise ValueError(
            "Bit resolution does not fall within the supported range [1, 64]"
        )


def simple_adc(
    detector: Detector,
    data_type: Optional[Literal["uint8", "uint16", "uint32", "uint64"]] = None,
) -> None:
    """Apply simple Analog-to-Digital Converter (ADC) to the `signal` bucket from `Detector`.

    This model simulates the behaviour of an ADC based on `detector.characteristics.adc_bit_resolution`.

    Parameters
    ----------
    detector : Detector
        Pyxel Detector object.
    data_type : str
        The desired data-type for the Image array. The data-type must be an unsigned integer.
        If not provided, the data type is determined based on `detector.characteristics.adc_bit_resolution`.
        Valid values: 'uint8', 'uint16', 'uint32', 'uint64'
        Invalid values: 'int16', 'int32', 'int64', 'int', 'float'...
    """

    bit_resolution = detector.characteristics.adc_bit_resolution
    voltage_min, voltage_max = detector.characteristics.adc_voltage_range

    if data_type:
        try:
            d_type: np.dtype = np.dtype(data_type)
        except TypeError as ex:
            raise TypeError(
                "Can not locate the type defined as `data_type` argument in yaml file."
            ) from ex

        if not issubclass(d_type.type, np.integer):
            raise TypeError("Expecting a signed/unsigned integer.")
    else:
        d_type = _get_dtype(bit_resolution)

    detector.image.array = apply_simple_adc(
        signal=detector.signal.array,
        bit_resolution=bit_resolution,
        voltage_min=voltage_min,
        voltage_max=voltage_max,
        dtype=d_type,
    )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Model to compute Mean-Variance."""
from typing import TYPE_CHECKING, Literal, Optional, Union

import xarray as xr
from datatree import DataTree

from pyxel.detectors import Detector

if TYPE_CHECKING:
    from pyxel.data_structure import Image, Photon, Pixel, Signal


def compute_mean_variance(data_array: xr.DataArray) -> xr.DataArray:
    """Compute mean-variance into a DataArray.

    Parameters
    ----------
    data_array : DataArray

    Returns
    -------
    DataArray

    Examples
    --------
    >>> data_array
    <xarray.DataArray 'image' (y: 100, x: 100)>
    array([[10406, 10409, 10472, ..., 10394, 10302, 10400],
           [10430, 10473, 10443, ..., 10427, 10446, 10452],
           [10456, 10524, 10479, ..., 10502, 10435, 10499],
           ...,
           [10381, 10385, 10552, ..., 10471, 10443, 10468],
           [10381, 10396, 10381, ..., 10472, 10380, 10509],
           [10455, 10431, 10382, ..., 10405, 10429, 10491]], dtype=uint32)
    Coordinates:
      * y        (y) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
      * x        (x) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
    Attributes:
        units:      adu
        long_name:  Image

    >>> mean_variance = compute_mean_variance(data_array)
    >>> mean_variance
    <xarray.DataArray 'variance' (mean: 1)>
    array([2130.03464796])
    Coordinates:
      * mean     (mean) float64 1.043e+04
    Attributes:
        units:    adu^2
    """
    mean = data_array.mean(dim=["y", "x"])
    variance = data_array.var(dim=["y", "x"])

    mean_variance: xr.DataArray = variance.expand_dims(mean=[mean])
    mean_variance.name = "variance"

    if unit := data_array.attrs.get("units"):
        mean_variance["mean"].attrs["units"] = unit
        mean_variance.attrs["units"] = f"{unit}^2"

    return mean_variance


def mean_variance(
    detector: Detector,
    data_structure: Literal["pixel", "photon", "image", "signal"] = "image",
    name: Optional[str] = None,
) -> None:
    """Compute mean-variance and store it in '.data' bucket.

    Parameters
    ----------
    detector : Detector
    data_structure : 'pixel', 'photon', 'image' or 'signal'
        Data bucket to use for the linear regression.
    name : str, optional
        Name to use for the result.

    Examples
    --------
    >>> import pyxel
    >>> config = pyxel.load("exposure_mode.yaml")

    Run exposure mode with 'data_processing/mean_variance' model

    >>> data_tree = pyxel.run_mode(
    ...     mode=config.exposure,
    ...     detector=config.detector,
    ...     pipeline=config.pipeline,
    ... )

    Get results

    >>> data_tree["/data/mean_variance"]
    DataTree('mean_variance', parent="data")
    └── DataTree('image')
            Dimensions:   (mean: 19)
            Coordinates:
              * mean      (mean) float64 5.723e+03 1.144e+04 ... 5.238e+04 5.238e+04
            Data variables:
                variance  (mean) float64 3.238e+06 1.294e+07 2.91e+07 ... 4.03e+05 3.778e+05

    >>> mean_variance = data_tree["/data/mean_variance/image/variance"]
    >>> mean_variance
    <xarray.DataArray 'variance' (mean: 19)>
    array([ 3238372.98476575, 12940428.69349581, 29101219.32832065,
           51711636.17586413, 68442386.58453502, 73522594.46113221,
           70591532.43505114, 54799694.83319437, 36351341.15700997,
           19577907.89405003,  7877859.3907925 ,  2284048.60403896,
             839388.11541348,   605169.50892288,   527554.54598688,
             474694.21759011,   434408.74359875,   402970.36722454,
             377784.5672024 ])
    Coordinates:
      * mean     (mean) float64 5.723e+03 1.144e+04 ... 5.238e+04 5.238e+04
    Attributes:
        units:    adu^2

    Display mean-variance plot

    >>> mean_variance.plot()

    .. figure:: _static/mean_variance_plot.png
        :scale: 70%
        :alt: Mean-Variance plot
        :align: center
    """
    if name is None:
        name = data_structure

    # Extract data from 'detector'
    data_bucket: Union[Pixel, Photon, Image, Signal] = getattr(detector, data_structure)
    data_array: xr.DataArray = data_bucket.to_xarray(dtype=float)

    # Get Mean-Variance data
    mean_variance: xr.DataArray = compute_mean_variance(data_array)

    parent: str = "/mean_variance"
    parent_partial: str = f"{parent}/partial"
    key: str = f"{parent}/{name}"
    key_partial: str = f"{parent_partial}/{name}"

    # TODO: Use 'detector.data.get(...)'
    try:
        _ = detector.data[key_partial]
    except KeyError:
        has_key_partial = False
    else:
        has_key_partial = True

    if not has_key_partial:
        data_tree: DataTree = DataTree(mean_variance)
    else:
        # Concatenate data
        data_tree = detector.data[key_partial].combine_first(mean_variance)  # type: ignore

    if detector.pipeline_count == (detector.num_steps - 1):
        detector.data[parent_partial].orphan()
        detector.data[key] = data_tree.sortby("mean")
    else:
        detector.data[key_partial] = data_tree



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


from collections.abc import Sequence
from pathlib import Path

import numpy as np
import pytest

from pyxel.detectors import (
    CCD,
    CMOS,
    CCDGeometry,
    Characteristics,
    CMOSGeometry,
    Environment,
    ReadoutProperties,
)
from pyxel.models.charge_collection import persistence


@pytest.fixture
def ccd_5x5() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=5,
            col=5,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )
    detector._readout_properties = ReadoutProperties(times=[1.0])
    return detector


@pytest.fixture
def cmos_5x5() -> CMOS:
    """Create a valid CCD detector."""
    detector = CMOS(
        geometry=CMOSGeometry(
            row=5,
            col=5,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )
    detector._readout_properties = ReadoutProperties(times=[1.0])
    return detector


@pytest.fixture
def valid_density_map_path(
    tmp_path: Path,
) -> str:
    """Create valid 2D file on a temporary folder."""
    data_2d = np.ones((5, 5)) * 0.01

    final_path = f"{tmp_path}/densities.npy"
    np.save(final_path, arr=data_2d)

    return final_path


@pytest.fixture
def valid_capacity_map_path(
    tmp_path: Path,
) -> str:
    """Create valid 2D file on a temporary folder."""
    data_2d = np.ones((5, 5)) * 100.0

    final_path = f"{tmp_path}/capacities.npy"
    np.save(final_path, arr=data_2d)

    return final_path


@pytest.fixture
def invalid_density_map_path(
    tmp_path: Path,
) -> str:
    """Create valid 2D file on a temporary folder."""

    data_2d = np.array(
        [
            [1.0, 1.0, 0.5, 1.0, 1.0],
            [1.0, 2.0, 1.0, 1.0, 0.0],
            [1.0, 1.0, 1.0, 1.0, 0.0],
            [1.0, 1.0, 1.0, 2.0, 1.0],
            [0.5, 1.0, 1.0, 1.0, 1.0],
        ]
    )

    final_path = f"{tmp_path}/invalid_densities.npy"
    np.save(final_path, arr=data_2d)

    return final_path


@pytest.mark.parametrize(
    "trap_time_constants, trap_proportions",
    [
        pytest.param([1.0, 10.0], [0.4, 0.6], id="valid"),
    ],
)
def test_persistence_valid(
    cmos_5x5: CMOS,
    trap_time_constants: Sequence[float],
    trap_proportions: Sequence[float],
    valid_density_map_path: str,
    valid_capacity_map_path: str,
):
    detector = cmos_5x5

    assert not detector.has_persistence()

    persistence(
        detector=detector,
        trap_time_constants=trap_time_constants,
        trap_proportions=trap_proportions,
        trap_densities_filename=valid_density_map_path,
        trap_capacities_filename=valid_capacity_map_path,
    )

    assert detector.has_persistence()
    assert len(detector.persistence.trap_list) == len(trap_time_constants)


@pytest.mark.parametrize(
    "trap_time_constants, trap_proportions, exp_error, exp_msg",
    [
        pytest.param(
            [],
            [],
            ValueError,
            "Expecting at least one 'trap_time_constants' and 'trap_proportions'",
            id="no elements",
        ),
        pytest.param(
            [1.0],
            [0.1, 0.1],
            ValueError,
            "Expecting same number of elements for parameters",
            id="not same number of elements",
        ),
    ],
)
def test_persistence_invalid(
    cmos_5x5: CMOS,
    trap_time_constants: Sequence[float],
    trap_proportions: Sequence[float],
    valid_density_map_path: str,
    valid_capacity_map_path: str,
    exp_error,
    exp_msg,
):
    detector = cmos_5x5

    with pytest.raises(exp_error, match=exp_msg):
        persistence(
            detector=detector,
            trap_time_constants=trap_time_constants,
            trap_proportions=trap_proportions,
            trap_densities_filename=valid_density_map_path,
            trap_capacities_filename=valid_capacity_map_path,
        )


def test_persistence_with_ccd(
    ccd_5x5: CCD, valid_density_map_path: str, valid_capacity_map_path: str
):
    """Test model 'persistence' with a `CCD` detector."""
    detector = ccd_5x5

    with pytest.raises(TypeError, match="Expecting a CMOS object for detector."):
        persistence(
            detector=detector,
            trap_time_constants=[1.0, 10.0],
            trap_proportions=[0.4, 0.6],
            trap_densities_filename=valid_density_map_path,
            trap_capacities_filename=valid_capacity_map_path,
        )


def test_persistence_with_invalid_density_map(
    cmos_5x5: CMOS, invalid_density_map_path: str
):
    """Test model 'persistence' with an invalid density map."""
    detector = cmos_5x5

    with pytest.raises(
        ValueError, match="Trap density map values not between 0 and 1."
    ):
        persistence(
            detector=detector,
            trap_time_constants=[1.0, 10.0],
            trap_proportions=[0.4, 0.6],
            trap_densities_filename=invalid_density_map_path,
        )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


"""Utility functions for mercury cadmium telluride alloy material properties."""
import numpy as np


class Mercadtel:
    """Useful functions to compute properties of mercury cadmium telluride alloy."""


def lambda_e(lambda_cutoff: float) -> float:
    """Compute lambda_e.

    Parameters
    ----------
    lambda_cutoff: int
        Cut-off wavelength of the detector.

    Returns
    -------
    float
    """
    lambda_scale = 0.200847413  # um
    lambda_threshold = 4.635136423  # um
    pwr = 0.544071282
    if lambda_cutoff < lambda_threshold:
        le = lambda_cutoff / (
            1
            - ((lambda_scale / lambda_cutoff) - (lambda_scale / lambda_threshold))
            ** pwr
        )
    else:
        le = lambda_cutoff
    return le


def energy_bandgap(cut_off: float):
    """Compute the energy of the MCT bandgap as a function of its cut-off wavelength.

    Parameters
    ----------
    cut_off: float
        the cut_off wavelength in micrometer

    Returns
    -------
    float
        the bandgap energy in eV
    """

    e_c = 1.24 / lambda_e(cut_off)
    return e_c


def eg_hansen_inverse(eg, temperature):
    """Return the cadmium fraction as a function of the temperature and the bandgap energy.

    Ref Biblio: G. L. HANSEN, J. L. SCHMIT, T. N. CASSELMAN, J. Appl. Phys., 53 (1982), p. 7099-7101

    Parameters
    ----------
    eg: float
        gap energy [eV]
    temperature: float
        Temperature [K]

    Returns
    -------
    float
        MCT Cd concentration
    """
    aa = 0.302
    bb = 1.93
    cc = 0.535 / 1000
    dd = 0.810
    ee = 0.832
    xx = (
        (
            (
                np.sqrt(
                    -32.0 * cc**3.0 * ee * temperature**3
                    - (
                        -27.0 * cc**2.0 * ee**2
                        + (36.0 * cc**2.0 * dd - 48.0 * bb * cc**2) * ee
                        + 4.0 * cc**2.0 * dd**2
                    )
                    * temperature**2
                    - (
                        (54.0 * cc * ee**2 - 36.0 * cc * dd * ee) * eg
                        + 54.0 * aa * cc * ee**2
                        + ((-18.0 * bb - 36.0 * aa) * cc * dd + 24.0 * bb**2.0 * cc)
                        * ee
                        + 4.0 * cc * dd**3
                        - 4.0 * bb * cc * dd**2
                    )
                    * temperature
                    + 27.0 * ee**2.0 * eg**2
                    - (-54.0 * aa * ee**2 + 18.0 * bb * dd * ee - 4.0 * dd**3) * eg
                    + 27.0 * aa**2.0 * ee**2
                    - (18.0 * aa * bb * dd - 4.0 * bb**3) * ee
                    + 4.0 * aa * dd**3
                    - bb**2.0 * dd**2
                )
            )
            / (6.0 * np.sqrt(3) * ee**2)
            + (
                cc * (18.0 * dd * ee * temperature - 27.0 * ee**2.0 * temperature)
                + 27.0 * ee**2.0 * eg
                + 27.0 * aa * ee**2
                - 9.0 * bb * dd * ee
                + 2.0 * dd**3
            )
            / (54.0 * ee**3)
        )
        ** (1.0 / 3)
        - (-6.0 * cc * ee * temperature + 3.0 * bb * ee - dd**2)
        / (
            9.0
            * ee**2.0
            * (
                (
                    np.sqrt(
                        -32.0 * cc**3.0 * ee * temperature**3
                        - (
                            -27.0 * cc**2.0 * ee**2
                            + (36.0 * cc**2.0 * dd - 48.0 * bb * cc**2) * ee
                            + 4.0 * cc**2.0 * dd**2
                        )
                        * temperature**2
                        - (
                            (54.0 * cc * ee**2 - 36.0 * cc * dd * ee) * eg
                            + 54.0 * aa * cc * ee**2
                            + (
                                (-18.0 * bb - 36.0 * aa) * cc * dd
                                + 24.0 * bb**2.0 * cc
                            )
                            * ee
                            + 4.0 * cc * dd**3
                            - 4.0 * bb * cc * dd**2
                        )
                        * temperature
                        + 27.0 * ee**2.0 * eg**2
                        - (-54.0 * aa * ee**2 + 18.0 * bb * dd * ee - 4.0 * dd**3)
                        * eg
                        + 27.0 * aa**2.0 * ee**2
                        - (18.0 * aa * bb * dd - 4.0 * bb**3) * ee
                        + 4.0 * aa * dd**3
                        - bb**2.0 * dd**2
                    )
                )
                / (6.0 * np.sqrt(3) * ee**2)
                + (
                    cc * (18.0 * dd * ee * temperature - 27.0 * ee**2.0 * temperature)
                    + 27.0 * ee**2.0 * eg
                    + 27.0 * aa * ee**2
                    - 9.0 * bb * dd * ee
                    + 2.0 * dd**3
                )
                / (54.0 * ee**3)
            )
            ** (1.0 / 3)
        )
        + dd / (3.0 * ee)
    )
    return xx


def eg_hansen(x, temperature):
    """Compute the MCT Energy gap computation as a function of the Cd fraction.

    Ref Biblio: G. L. HANSEN, J. L. SCHMIT, T. N. CASSELMAN, J. Appl. Phys., 53 (1982), p. 7099-7101

    Parameters
    ----------
    x: float
    MCT Cd concentration (should be smaller than 1)
    temperature: float
        temperature [K]

    Returns
    -------
    float
         gap energy [eV]
    """
    eg = (
        -0.302
        + 1.93 * x
        - 0.810 * x**2
        + 0.832 * x**3
        + 0.535 * (1 - 2 * x) * temperature / 1000
    )
    return eg


def absorption_coefficient(e, x, temperature):
    """Compute the absorption coefficient of MCT alloy using a non-parabolic band model and an urbach tail.

    Ref: Moazzami, JEM 34(6) p773 (2005)

    Parameters
    ----------
    e: float
        photon energy [eV]
    x: float
        MCT Cd composition fraction (should be smaller than one)
    temperature: float
        temperature [K]

    Returns
    -------
    float
        the absorption coefficient
    """

    w = 4.96e-3
    k = (
        -20060
        + 115750 * x
        + 32.43 * temperature
        - 64170 * x**2
        + 0.43231 * temperature**2
        - 101.92 * x * temperature
    )
    n = 0.74487 - 0.44513 * x + 1e-4 * (7.99 - 7.57 * x) * temperature
    eg = eg_hansen(x, temperature)
    delta = (eg + w * (n - 1)) ** 2 + 4 * w * eg
    eo = (eg + w * (n - 1) + np.sqrt(delta)) / 2
    alphao = k * (eo - eg) ** n / eo
    alpha1 = alphao * np.exp((e - eo) / w)
    alpha2 = k * (e - eg) ** n / e
    if e < eo:
        alpha = alpha1
    else:
        alpha = alpha2
    return alpha


def density(x):
    """Compute the MCT density as a function of the cadmium fraction.

    Parameters
    ----------
    x: float
        the cadmium fraction

    Returns
    -------
    float
    the density in g/cm3
    """
    me_density = 5.43
    cd_density = 8.65
    te_density = 6.24
    mct_density = 0.5 * (x / cd_density + (1 - x) / me_density) + 0.5 / te_density
    mct_density = 1.0 / mct_density
    return mct_density



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from pyxel.exposure.readout import calculate_steps


@pytest.mark.parametrize(
    "times, start_time, exp_steps",
    [
        (np.array([1, 2, 4, 7, 10]), 0.0, np.array([1.0, 1.0, 2.0, 3.0, 3.0])),
        (np.array([1, 2, 4, 7, 10]), 0.5, np.array([0.5, 1.0, 2.0, 3.0, 3.0])),
    ],
)
def test_calculate_steps(times, start_time, exp_steps):
    """Test function 'calculate_steps'."""
    steps = calculate_steps(times=times, start_time=start_time)

    assert_array_equal(steps, exp_steps)



#   Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#   This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#   is part of this Pyxel package. No part of the package, including
#   this file, may be copied, modified, propagated, or distributed except according to
#   the terms contained in the file ‘LICENCE.txt’.

import pytest

from pyxel.evaluator import eval_range


@pytest.mark.parametrize(
    "expr, exp",
    [
        ("[1,2,3]", [1, 2, 3]),
        ([1, 2, 3], [1, 2, 3]),
        ("numpy.array([1,2,3])", [1, 2, 3]),
        ("numpy.array([1,2,3], dtype=float)", [1.0, 2.0, 3.0]),
        ("numpy.array([1,2,3], dtype=numpy.float64)", [1.0, 2.0, 3.0]),
        ("numpy.array([1,2,3], dtype=int)", [1, 2, 3]),
        ("numpy.array([1,2,3], dtype=numpy.int64)", [1, 2, 3]),
        ("numpy.arange(0, 5, 2)", [0, 2, 4]),
        ("_", ["_"]),
    ],
)
def test_eval_range(expr: str, exp):
    """Test function 'eval_range'."""
    value = eval_range(expr)
    assert value == exp


@pytest.mark.parametrize(
    "expr, exp_exc, exp_msg",
    [
        (
            "numpy.array([1,2,3], dtype=numpy.float32)",
            NotImplementedError,
            "numpy data type is not a float or int",
        ),
        (
            "numpy.array([1,2,3], dtype=numpy.uint64)",
            NotImplementedError,
            "numpy data type is not a float or int",
        ),
        ("np.array([1,2,3])", NameError, r"name \'np\' is not defined"),
        (1, NotImplementedError, ""),
    ],
)
def test_eval_range_failing(expr: str, exp_exc: Exception, exp_msg: str):
    """Test function 'eval_range' with bad inputs."""
    with pytest.raises(exp_exc, match=exp_msg):
        _ = eval_range(expr)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""

# flake8: noqa
from .measurement import simple_measurement
from .readout_noise import (
    output_node_noise,
    output_node_noise_cmos,
    readout_noise_saphira,
)
from .nghxrg.nghxrg import nghxrg
from .linearity import (
    output_node_linearity_poly,
    simple_physical_non_linearity,
    physical_non_linearity,
    physical_non_linearity_with_saturation,
)
from .offset import dc_offset, output_pixel_reset_voltage_apd
from .reset_noise import ktc_noise



#   Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#  #
#   This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#   is part of this Pyxel package. No part of the package, including
#   this file, may be copied, modified, propagated, or distributed except according to
#   the terms contained in the file ‘LICENCE.txt’.

import pytest
import xarray as xr
from pytest_mock import MockerFixture  # pip install pytest-mock

from pyxel.data_structure import Scene
from pyxel.models.photon_collection.aperture import extract_wavelength


def test_extract_wavelength_error():
    with pytest.raises(
        ValueError,
        match="First input in wavelength_band needs to be smaller than the second.",
    ):
        extract_wavelength(scene=..., wavelength_band=[900, 500])


@pytest.fixture
def scene_dataset() -> xr.Dataset:
    dct = {
        "coords": {
            "ref": {"dims": ("ref",), "attrs": {}, "data": [0, 1, 2]},
            "wavelength": {
                "dims": ("wavelength",),
                "attrs": {"units": "nm"},
                "data": [336.0, 338.0, 340.0, 342.0],
            },
        },
        "attrs": {},
        "dims": {"ref": 3, "wavelength": 4},
        "data_vars": {
            "x": {
                "dims": ("ref",),
                "attrs": {"units": "arcsec"},
                "data": [205732.81147230256, 205832.50867371075, 206010.7213446728],
            },
            "y": {
                "dims": ("ref",),
                "attrs": {"units": "arcsec"},
                "data": [85748.89015925185, 85802.09354969644, 85961.12201291585],
            },
            "weight": {
                "dims": ("ref",),
                "attrs": {"units": "mag"},
                "data": [11.489056587219238, 14.131349563598633, 15.217577934265137],
            },
            "flux": {
                "dims": ("ref", "wavelength"),
                "attrs": {"units": "ph / (cm2 nm s)"},
                "data": [
                    [
                        0.03769071172453367,
                        0.041374086069586175,
                        0.03988154035070513,
                        0.0362458369803364,
                    ],
                    [
                        0.01151902543689362,
                        0.010221036617896213,
                        0.009752581546174546,
                        0.009850105680742818,
                    ],
                    [
                        0.0019267151902375244,
                        0.0017953813076332019,
                        0.0016775406524146613,
                        0.001506762466011516,
                    ],
                ],
            },
        },
    }
    return xr.Dataset.from_dict(dct)


@pytest.fixture
def dummy_scene(mocker: MockerFixture, scene_dataset: xr.Dataset):
    scene = Scene()

    mocker.patch.object(scene, "to_xarray", return_value=scene_dataset)
    return scene


def test_extract_wavelength(dummy_scene: Scene, scene_dataset: xr.Dataset):
    """..."""
    ds = extract_wavelength(scene=dummy_scene, wavelength_band=[336, 340])

    exp_ds = scene_dataset.copy().sel(wavelength=slice(336, 340))

    xr.testing.assert_equal(ds, exp_ds)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""TBW."""

from collections.abc import Iterator, Sequence
from typing import Optional, Union

import numpy as np

from pyxel import load_table
from pyxel.evaluator import eval_range


class Readout:
    """Readout configuration that contains parameters for a time-domain process.

    This class supports both linear and custom sampling times, which can be loaded from a
    file or defined directly.

    Parameters
    ----------
    times : Sequence[Number] or str, optional
        A sequence of numeric values or a string representing the sampling times for the
        readout simulation. It can be a list, tuple, or other iterable containing numerical
        values (e.g. [1.0, 2.0, 3.0]). Alternatively, it can be a string representing a range of numbers using
        Python's syntax (e.g. 'numpy.linspace(0, 10, 100)').
        If not provided, the default readout/exposure time is 1 second.

    times_from_file : str, optional
        A string specifying the path to a file containing the sampling times for the readout.
        Parameters ``times`` and ``times_from_file`` cannot be provided at the same time.

    start_time : float, optional. Default: 0.0
        A float representing the starting time of the readout simulation.
        The readout time(s) should be greater that this ``start_time``.

    non_destructive : bool, optional. Default: False
        A boolean flag indicating whether the readout simulation is non-destructive.
        If set to ``True``, the readout process will not modify the underlying data.

    Raises
    ------
    ValueError
        Raises if both ``times`` and ``times_from_file`` parameters are provided or if neither
        of them is specified.
        Raises if the readout times are not strictly increasing or the first readout time is zero.
        Raises if the ``start_time`` parameter is greater than or equal to the first readout time.

    Examples
    --------
    Example 1: Using linearly spaced readout times

    >>> readout1 = Readout(times=[1, 2, 3, 4, 5], start_time=0.5, non_destructive=False)

    Example 2: Loading readout times from a file

    >>> readout2 = Readout(
    ...     times_from_file="readout_times.csv",
    ...     start_time=0.0,
    ...     non_destructive=True,
    ... )
    """

    def __init__(
        self,
        times: Union[Sequence, str, None] = None,
        times_from_file: Optional[str] = None,
        start_time: float = 0.0,
        non_destructive: bool = False,
    ):
        self._time_domain_simulation = True

        if times is not None and times_from_file is not None:
            raise ValueError("Both times and times_from_file specified. Choose one.")
        elif times is times_from_file is None:
            # by convention default readout/exposure time is 1 second
            self._times = np.array([1])
            self._time_domain_simulation = False
        elif times_from_file:
            self._times = load_table(times_from_file).to_numpy(dtype=float).flatten()
        elif times:
            self._times = np.array(eval_range(times), dtype=float)
        else:
            raise ValueError("Sampling times not specified.")

        if self._times[0] == 0:
            raise ValueError("Readout times should be non-zero values.")
        elif start_time >= self._times[0]:
            raise ValueError("Readout times should be greater than start time.")

        if not np.all(np.diff(self._times) > 0):
            raise ValueError("Readout times must be strictly increasing")

        self._non_destructive = non_destructive

        self._times_linear: bool = True
        self._start_time: float = start_time
        self._steps: np.ndarray = np.array([])
        self._num_steps: int = 0
        self._set_steps()

    def __repr__(self) -> str:
        cls_name: str = self.__class__.__name__
        return f"{cls_name}<num_steps={self._num_steps}>"

    def _set_steps(self) -> None:
        """TBW."""
        self._steps = calculate_steps(times=self._times, start_time=self._start_time)
        self._times_linear = bool(np.all(self._steps == self._steps[0]))
        self._num_steps = len(self._times)

    def time_step_it(self) -> Iterator[tuple[float, float]]:
        """TBW."""
        return zip(self._times, self._steps)

    @property
    def start_time(self) -> float:
        """Return start time of the readout."""
        return self._start_time

    @start_time.setter
    def start_time(self, value: float) -> None:
        """Set start time of the readout."""
        if value >= self._times[0]:
            raise ValueError("Readout times should be greater than start time.")
        self._start_time = value
        self._set_steps()

    @property
    def times(self) -> np.ndarray:
        """Return readout times."""
        return self._times

    @times.setter
    def times(self, value: Union[Sequence, np.ndarray]) -> None:
        """Set readout times.

        Parameters
        ----------
        value
        """
        if value[0] == 0:
            raise ValueError("Readout times should be non-zero values.")
        elif self._start_time >= value[0]:
            raise ValueError("Readout times should be greater than start time.")
        self._times = np.array(value)
        self._set_steps()

    @property
    def time_domain_simulation(self) -> bool:
        """TBW."""
        return self._time_domain_simulation

    @property
    def steps(self) -> np.ndarray:
        """TBW."""
        return self._steps

    @property
    def non_destructive(self) -> bool:
        """Get non-destructive readout mode."""
        return self._non_destructive

    @non_destructive.setter
    def non_destructive(self, value: bool) -> None:
        """Set non-destructive mode."""
        self._non_destructive = value


def calculate_steps(times: np.ndarray, start_time: float) -> np.ndarray:
    """Calculate time steps between consecutive time points.

    Parameters
    ----------
    times : ndarray
        A numpy array representing the time points for which the time steps will be calculated.
        The array should contain numerical values and must be one-dimensional (1D).
    start_time : float
        A float representing the starting time for which the time steps will be computed.
        This value serves as the reference for the first time step.

    Returns
    -------
    ndarray
        An array containing the time steps between consecutive time points in the ``times`` array.

    Examples
    --------
    >>> calculate_steps(times=np.array([1, 2, 4, 7, 10]), start_time=0.0)
    array([1., 1., 2., 3., 3.])

    >>> calculate_steps(times=np.array([1, 2, 4, 7, 10]), start_time=0.5)
    array([0.5, 1., 2., 3., 3.])
    """
    steps = np.diff(
        np.concatenate((np.array([start_time]), times), axis=0),
        axis=0,
    )

    return steps



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

from pathlib import Path

import pytest

import pyxel
from pyxel.observation import Observation
from pyxel.pipelines import Processor


@pytest.mark.parametrize(
    "filename, exp_exc, exp_msg",
    [
        pytest.param(
            Path("observation_missing_param.yaml"),
            KeyError,
            "Missing parameter: 'pipeline.charge_generation.dark_current.arguments.temperature'",
            id="observation_missing_param",
        ),
        pytest.param(
            Path("observation_not_enabled.yaml"),
            ValueError,
            "The 'pipeline.charge_generation.dark_current' model referenced in Observation "
            "configuration has not been enabled",
            id="observation_not_enabled",
        ),
        pytest.param(
            Path("deprecated_observation_missing_param.yaml"),
            KeyError,
            "Missing parameter: 'pipeline.charge_generation.dark_current.arguments.temperature'",
            id="deprecated_observation_missing_param",
        ),
        pytest.param(
            Path("deprecated_observation_not_enabled.yaml"),
            ValueError,
            "The 'pipeline.charge_generation.dark_current' model referenced in Observation "
            "configuration has not been enabled",
            id="deprecated_observation_not_enabled",
        ),
    ],
)
def test_validate_steps(filename: Path, exp_exc, exp_msg: str):
    """Test method 'Observation.validate_steps'."""
    folder = Path("tests/pipelines/observation/config").resolve()
    assert folder.exists()

    filename = folder / filename
    assert filename.exists()

    # Read the configuration file
    cfg = pyxel.load(filename)

    # Get running mode
    observation = cfg.running_mode
    assert isinstance(observation, Observation)

    # Create a 'Processor' object
    processor = Processor(detector=cfg.detector, pipeline=cfg.pipeline)

    with pytest.raises(exp_exc, match=exp_msg):
        observation.validate_steps(processor)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


from dataclasses import dataclass
from typing import Literal

import numpy as np
import pandas as pd
import pytest

from pyxel.data_structure import Charge
from pyxel.detectors import Geometry


@pytest.fixture
def geo() -> Geometry:
    return Geometry(
        row=10, col=10, total_thickness=10.0, pixel_horz_size=1.0, pixel_vert_size=1.0
    )


def test_empty_charge(geo: Geometry):
    """Test an empty `Charge` object."""
    charge = Charge(geo=geo)

    assert charge.nextid == 0

    pd.testing.assert_frame_equal(
        charge.frame,
        pd.DataFrame(
            data=None,
            columns=(
                "charge",
                "number",
                "init_energy",
                "energy",
                "init_pos_ver",
                "init_pos_hor",
                "init_pos_z",
                "position_ver",
                "position_hor",
                "position_z",
                "velocity_ver",
                "velocity_hor",
                "velocity_z",
            ),
            dtype=float,
        ),
    )


def test_add_one_charge(geo: Geometry):
    """Test with one charge."""
    charge = Charge(geo=geo)

    charge.add_charge(
        particle_type="e",
        particles_per_cluster=np.array([1]),
        init_energy=np.array([0.1]),
        init_ver_position=np.array([1.1]),
        init_hor_position=np.array([2.2]),
        init_z_position=np.array([3.3]),
        init_ver_velocity=np.array([4.4]),
        init_hor_velocity=np.array([-5.5]),
        init_z_velocity=np.array([6.6]),
    )

    exp_df_charges = pd.DataFrame(
        {
            "charge": np.array([-1]),
            "number": np.array([1]),
            "init_energy": np.array([0.1]),
            "energy": np.array([0.1]),
            "init_pos_ver": np.array([1.1]),
            "init_pos_hor": np.array([2.2]),
            "init_pos_z": np.array([3.3]),
            "position_ver": np.array([1.1]),
            "position_hor": np.array([2.2]),
            "position_z": np.array([3.3]),
            "velocity_ver": np.array([4.4]),
            "velocity_hor": np.array([-5.5]),
            "velocity_z": np.array([6.6]),
        }
    )

    # ensure that the integer types are consistent
    exp_df_charges["charge"] = exp_df_charges["charge"].astype(np.int64)
    exp_df_charges["number"] = exp_df_charges["number"].astype(np.int64)
    charge.frame["charge"] = charge.frame["charge"].astype(np.int64)
    charge.frame["number"] = charge.frame["number"].astype(np.int64)

    assert charge.nextid == 1
    pd.testing.assert_frame_equal(charge.frame, exp_df_charges)


def test_add_one_hole(geo: Geometry):
    """Test with one hole."""
    charge = Charge(geo=geo)

    charge.add_charge(
        particle_type="h",
        particles_per_cluster=np.array([1]),
        init_energy=np.array([0.1]),
        init_ver_position=np.array([1.1]),
        init_hor_position=np.array([2.2]),
        init_z_position=np.array([3.3]),
        init_ver_velocity=np.array([4.4]),
        init_hor_velocity=np.array([-5.5]),
        init_z_velocity=np.array([6.6]),
    )

    exp_df_charges = pd.DataFrame(
        {
            "charge": [1],
            "number": [1],
            "init_energy": [0.1],
            "energy": [0.1],
            "init_pos_ver": [1.1],
            "init_pos_hor": [2.2],
            "init_pos_z": [3.3],
            "position_ver": [1.1],
            "position_hor": [2.2],
            "position_z": [3.3],
            "velocity_ver": [4.4],
            "velocity_hor": [-5.5],
            "velocity_z": [6.6],
        }
    )

    # ensure that the integer types are consistent
    exp_df_charges["charge"] = exp_df_charges["charge"].astype(np.int64)
    exp_df_charges["number"] = exp_df_charges["number"].astype(np.int64)
    charge.frame["charge"] = charge.frame["charge"].astype(np.int64)
    charge.frame["number"] = charge.frame["number"].astype(np.int64)

    assert charge.nextid == 1
    pd.testing.assert_frame_equal(charge.frame, exp_df_charges)


@dataclass
class ChargeInfo:
    """Define one charge.

    Only for testing.
    """

    particle_type: Literal["e", "h"]
    particles_per_cluster: np.ndarray
    init_energy: np.ndarray
    init_ver_position: np.ndarray
    init_hor_position: np.ndarray
    init_z_position: np.ndarray
    init_ver_velocity: np.ndarray
    init_hor_velocity: np.ndarray
    init_z_velocity: np.ndarray


@pytest.mark.parametrize(
    "param_name",
    [
        "particles_per_cluster",
        "init_energy",
        "init_ver_position",
        "init_hor_position",
        "init_z_position",
        "init_ver_velocity",
        "init_hor_velocity",
        "init_z_velocity",
    ],
)
@pytest.mark.parametrize(
    "new_value, exp_error",
    [
        pytest.param(
            np.array([]),
            r"List arguments have different lengths",
            id="Too few values",
        ),
        pytest.param(
            np.array([1.0, 2.0]),
            r"List arguments have different lengths",
            id="Too much values",
        ),
        pytest.param(
            np.array([[1.0]]),
            r"List arguments must have only one dimension",
            id="More than one dimension",
        ),
    ],
)
def test_invalid_add_charge(param_name: str, geo: Geometry, new_value, exp_error):
    """Test method `Charge.add_charge` with invalid parameters."""
    charge = Charge(geo=geo)

    # Create valid parameters
    params = ChargeInfo(
        particle_type="e",
        particles_per_cluster=np.array([1]),
        init_energy=np.array([0.1]),
        init_ver_position=np.array([1.1]),
        init_hor_position=np.array([2.2]),
        init_z_position=np.array([3.3]),
        init_ver_velocity=np.array([4.4]),
        init_hor_velocity=np.array([-5.5]),
        init_z_velocity=np.array([6.6]),
    )

    # Add the valid parameters
    charge.add_charge(
        particle_type=params.particle_type,
        particles_per_cluster=params.particles_per_cluster,
        init_energy=params.init_energy,
        init_ver_position=params.init_ver_position,
        init_hor_position=params.init_hor_position,
        init_z_position=params.init_z_position,
        init_ver_velocity=params.init_ver_velocity,
        init_hor_velocity=params.init_hor_velocity,
        init_z_velocity=params.init_z_velocity,
    )

    # Add an invalid parameter
    # This is equivalent to 'params.param_name = new_value'
    setattr(params, param_name, new_value)

    with pytest.raises(ValueError, match=exp_error):
        charge.add_charge(
            particle_type=params.particle_type,
            particles_per_cluster=params.particles_per_cluster,
            init_energy=params.init_energy,
            init_ver_position=params.init_ver_position,
            init_hor_position=params.init_hor_position,
            init_z_position=params.init_z_position,
            init_ver_velocity=params.init_ver_velocity,
            init_hor_velocity=params.init_hor_velocity,
            init_z_velocity=params.init_z_velocity,
        )


@pytest.mark.parametrize("particle_type", ["E", "H", "electron", "hole"])
def test_invalid_particle_type(geo: Geometry, particle_type: Literal["e", "h"]):
    """Test method `Charge.add_charge` with invalid 'particle_type."""
    charge = Charge(geo=geo)

    with pytest.raises(
        ValueError, match="Given charged particle type can not be simulated"
    ):
        charge.add_charge(
            particle_type=particle_type,
            particles_per_cluster=np.array([1]),
            init_energy=np.array([0.1]),
            init_ver_position=np.array([1.1]),
            init_hor_position=np.array([2.2]),
            init_z_position=np.array([3.3]),
            init_ver_velocity=np.array([4.4]),
            init_hor_velocity=np.array([-5.5]),
            init_z_velocity=np.array([6.6]),
        )


def test_add_two_charges(geo: Geometry):
    """Test when adding one charges in one time."""
    charge = Charge(geo=geo)

    charge.add_charge(
        particle_type="e",
        particles_per_cluster=np.array([1, 2]),
        init_energy=np.array([0.1, 0.2]),
        init_ver_position=np.array([1.11, 1.12]),
        init_hor_position=np.array([2.21, 2.22]),
        init_z_position=np.array([3.31, 3.32]),
        init_ver_velocity=np.array([4.41, 4.42]),
        init_hor_velocity=np.array([-5.51, 5.52]),
        init_z_velocity=np.array([6.61, 6.62]),
    )

    exp_df_charges = pd.DataFrame(
        {
            "charge": [-1, -1],
            "number": [1, 2],
            "init_energy": [0.1, 0.2],
            "energy": [0.1, 0.2],
            "init_pos_ver": [1.11, 1.12],
            "init_pos_hor": [2.21, 2.22],
            "init_pos_z": [3.31, 3.32],
            "position_ver": [1.11, 1.12],
            "position_hor": [2.21, 2.22],
            "position_z": [3.31, 3.32],
            "velocity_ver": [4.41, 4.42],
            "velocity_hor": [-5.51, 5.52],
            "velocity_z": [6.61, 6.62],
        }
    )

    # ensure that the integer types are consistent
    exp_df_charges["charge"] = exp_df_charges["charge"].astype(np.int64)
    exp_df_charges["number"] = exp_df_charges["number"].astype(np.int64)
    charge.frame["charge"] = charge.frame["charge"].astype(np.int64)
    charge.frame["number"] = charge.frame["number"].astype(np.int64)

    assert charge.nextid == 2
    pd.testing.assert_frame_equal(charge.frame, exp_df_charges)


def test_add_two_charges_one_hole(geo: Geometry):
    """Test when adding two charges and then one hole."""
    charge = Charge(geo=geo)

    # Add 2 charges
    charge.add_charge(
        particle_type="e",
        particles_per_cluster=np.array([1, 2]),
        init_energy=np.array([0.1, 0.2]),
        init_ver_position=np.array([1.11, 1.12]),
        init_hor_position=np.array([2.21, 2.22]),
        init_z_position=np.array([3.31, 3.32]),
        init_ver_velocity=np.array([4.41, 4.42]),
        init_hor_velocity=np.array([-5.51, 5.52]),
        init_z_velocity=np.array([6.61, 6.62]),
    )

    # Add 1 hole
    charge.add_charge(
        particle_type="h",
        particles_per_cluster=np.array([3]),
        init_energy=np.array([0.3]),
        init_ver_position=np.array([1.13]),
        init_hor_position=np.array([2.23]),
        init_z_position=np.array([3.33]),
        init_ver_velocity=np.array([4.43]),
        init_hor_velocity=np.array([5.53]),
        init_z_velocity=np.array([6.63]),
    )

    exp_df_charges = pd.DataFrame(
        {
            "charge": [-1, -1, 1],
            "number": [1, 2, 3],
            "init_energy": [0.1, 0.2, 0.3],
            "energy": [0.1, 0.2, 0.3],
            "init_pos_ver": [1.11, 1.12, 1.13],
            "init_pos_hor": [2.21, 2.22, 2.23],
            "init_pos_z": [3.31, 3.32, 3.33],
            "position_ver": [1.11, 1.12, 1.13],
            "position_hor": [2.21, 2.22, 2.23],
            "position_z": [3.31, 3.32, 3.33],
            "velocity_ver": [4.41, 4.42, 4.43],
            "velocity_hor": [-5.51, 5.52, 5.53],
            "velocity_z": [6.61, 6.62, 6.63],
        }
    )

    # ensure that the integer types are consistent
    exp_df_charges["charge"] = exp_df_charges["charge"].astype(np.int64)
    exp_df_charges["number"] = exp_df_charges["number"].astype(np.int64)
    charge.frame["charge"] = charge.frame["charge"].astype(np.int64)
    charge.frame["number"] = charge.frame["number"].astype(np.int64)

    assert charge.nextid == 3
    pd.testing.assert_frame_equal(charge.frame, exp_df_charges)


def test_invalid_add_charge_dataframe(geo: Geometry):
    """Test method `Charge.add_charge_dataframe` with an invalid input."""
    charge = Charge(geo=geo)

    df = pd.DataFrame(
        {
            "charge": [1.0],
            "number": [1.0],
            "init_energy": [0.1],
            "energy": [0.1],
            "init_pos_ver": [1.1],
            "init_pos_hor": [2.2],
            "init_pos_z": [3.3],
            "position_ver": [1.1],
            "position_hor": [2.2],
            "position_z": [3.3],
            "BAD_COLUMN": [4.4],  # Wrong column
            "velocity_hor": [-5.5],
            "velocity_z": [6.6],
        }
    )

    with pytest.raises(
        ValueError,
        match=r"Expected columns: 'charge', 'number', 'init_energy', 'energy', "
        r"'init_pos_ver', 'init_pos_hor', 'init_pos_z', "
        r"'position_ver', 'position_hor', 'position_z', "
        r"'velocity_ver', 'velocity_hor', 'velocity_z'",
    ):
        charge.add_charge_dataframe(new_charges=df)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

from pathlib import Path

import dask
import numpy as np
import pytest
import xarray as xr

import pyxel
from pyxel.configuration import Configuration
from pyxel.detectors import CCD
from pyxel.observation import Observation, ObservationResult, ParameterMode
from pyxel.pipelines import DetectionPipeline

dask.config.set(scheduler="single-threaded")


@pytest.mark.deprecated
@pytest.mark.parametrize("with_dask", [False, True])
def test_product_simple_old(with_dask: bool):
    """Test running mode 'observation'."""
    filename = Path(
        "tests/running_mode/data/deprecated_observation_product_simple.yaml"
    )
    assert filename.exists()

    cfg = pyxel.load(filename)
    assert isinstance(cfg, Configuration)

    observation = cfg.observation
    detector = cfg.detector
    pipeline = cfg.pipeline

    observation.with_dask = with_dask

    assert isinstance(observation, Observation)
    assert observation.parameter_mode == ParameterMode.Product

    assert isinstance(detector, CCD)
    assert isinstance(pipeline, DetectionPipeline)

    result = pyxel.observation_mode(
        observation=observation, detector=detector, pipeline=pipeline
    )
    assert isinstance(result, ObservationResult)

    # Check '.dataset'
    ds = result.dataset
    assert isinstance(ds, xr.Dataset)

    assert list(ds["image"].coords) == [
        "readout_time",
        "y",
        "x",
        "level",
        "quantum_efficiency",
    ]
    assert list(ds["image"].coords["readout_time"].values) == [1]
    assert list(ds["image"].coords["y"].values) == list(range(100))
    assert list(ds["image"].coords["x"].values) == list(range(100))
    assert list(ds["image"].coords["level"].values) == [10, 20, 30]
    assert list(ds["image"].coords["quantum_efficiency"].values) == [0.5, 0.8, 0.9]

    # Check '.parameters'
    parameters = result.parameters
    assert isinstance(parameters, xr.Dataset)

    exp_parameters = xr.Dataset(
        {
            "level": xr.DataArray(
                [10, 20, 30], dims="level_id", coords={"level_id": [0, 1, 2]}
            ),
            "quantum_efficiency": xr.DataArray(
                [0.5, 0.8, 0.9],
                dims="quantum_efficiency_id",
                coords={"quantum_efficiency_id": [0, 1, 2]},
            ),
        }
    )
    xr.testing.assert_equal(parameters, exp_parameters)

    # Check '.logs'
    logs = result.logs
    assert isinstance(logs, xr.Dataset)

    exp_logs = xr.Dataset(
        {
            "level": xr.DataArray([10, 10, 10, 20, 20, 20, 30, 30, 30], dims="id"),
            "quantum_efficiency": xr.DataArray(
                [0.5, 0.8, 0.9, 0.5, 0.8, 0.9, 0.5, 0.8, 0.9], dims="id"
            ),
        },
        coords={"id": [0, 1, 2, 3, 4, 5, 6, 7, 8]},
    )
    xr.testing.assert_equal(logs, exp_logs)


@pytest.mark.parametrize("with_dask", [False, True])
def test_product_simple(with_dask: bool):
    """Test running mode 'observation'."""
    filename = Path("tests/running_mode/data/observation_product_simple.yaml")
    assert filename.exists()

    cfg = pyxel.load(filename)
    assert isinstance(cfg, Configuration)

    observation = cfg.observation
    detector = cfg.detector
    pipeline = cfg.pipeline

    observation.with_dask = with_dask

    assert isinstance(observation, Observation)
    assert observation.parameter_mode == ParameterMode.Product

    assert isinstance(detector, CCD)
    assert isinstance(pipeline, DetectionPipeline)

    result = pyxel.observation_mode(
        observation=observation, detector=detector, pipeline=pipeline
    )
    assert isinstance(result, ObservationResult)

    # Check '.dataset'
    ds = result.dataset
    assert isinstance(ds, xr.Dataset)

    assert list(ds["image"].coords) == [
        "readout_time",
        "y",
        "x",
        "level",
        "quantum_efficiency",
    ]
    assert list(ds["image"].coords["readout_time"].values) == [1]
    assert list(ds["image"].coords["y"].values) == list(range(100))
    assert list(ds["image"].coords["x"].values) == list(range(100))
    assert list(ds["image"].coords["level"].values) == [10, 20, 30]
    assert list(ds["image"].coords["quantum_efficiency"].values) == [0.5, 0.8, 0.9]

    # Check '.parameters'
    parameters = result.parameters
    assert isinstance(parameters, xr.Dataset)

    exp_parameters = xr.Dataset(
        {
            "level": xr.DataArray(
                [10, 20, 30], dims="level_id", coords={"level_id": [0, 1, 2]}
            ),
            "quantum_efficiency": xr.DataArray(
                [0.5, 0.8, 0.9],
                dims="quantum_efficiency_id",
                coords={"quantum_efficiency_id": [0, 1, 2]},
            ),
        }
    )
    xr.testing.assert_equal(parameters, exp_parameters)

    # Check '.logs'
    logs = result.logs
    assert isinstance(logs, xr.Dataset)

    exp_logs = xr.Dataset(
        {
            "level": xr.DataArray([10, 10, 10, 20, 20, 20, 30, 30, 30], dims="id"),
            "quantum_efficiency": xr.DataArray(
                [0.5, 0.8, 0.9, 0.5, 0.8, 0.9, 0.5, 0.8, 0.9], dims="id"
            ),
        },
        coords={"id": [0, 1, 2, 3, 4, 5, 6, 7, 8]},
    )
    xr.testing.assert_equal(logs, exp_logs)


@pytest.mark.deprecated
@pytest.mark.parametrize("with_dask", [False, True])
def test_product_deprecated(with_dask: bool):
    """Test running mode 'observation'."""
    filename = Path("tests/running_mode/data/deprecated_observation_product.yaml")
    assert filename.exists()

    cfg = pyxel.load(filename)
    assert isinstance(cfg, Configuration)

    observation = cfg.observation
    detector = cfg.detector
    pipeline = cfg.pipeline

    observation.with_dask = with_dask

    assert isinstance(observation, Observation)
    assert observation.parameter_mode == ParameterMode.Product

    assert isinstance(detector, CCD)
    assert isinstance(pipeline, DetectionPipeline)

    result = pyxel.observation_mode(
        observation=observation, detector=detector, pipeline=pipeline
    )
    assert isinstance(result, ObservationResult)

    # Check '.dataset'
    ds = result.dataset
    assert isinstance(ds, xr.Dataset)

    assert list(ds["image"].coords) == [
        "readout_time",
        "y",
        "x",
        "illumination_uniform.level",
        "illumination_elliptic.level",
        "quantum_efficiency",
    ]
    assert list(ds["image"].coords["readout_time"].values) == [1]
    assert list(ds["image"].coords["y"].values) == list(range(100))
    assert list(ds["image"].coords["x"].values) == list(range(100))
    assert list(ds["image"].coords["illumination_uniform.level"].values) == [10, 20, 30]
    assert list(ds["image"].coords["illumination_elliptic.level"].values) == [
        100,
        200,
        300,
    ]
    assert list(ds["image"].coords["quantum_efficiency"].values) == [0.5, 0.8, 0.9]

    # Check '.parameters'
    parameters = result.parameters
    assert isinstance(parameters, xr.Dataset)

    exp_parameters = xr.Dataset(
        {
            "illumination_uniform.level": xr.DataArray(
                [10, 20, 30],
                dims="illumination_uniform.level_id",
                coords={"illumination_uniform.level_id": [0, 1, 2]},
            ),
            "illumination_elliptic.level": xr.DataArray(
                [100, 200, 300],
                dims="illumination_elliptic.level_id",
                coords={"illumination_elliptic.level_id": [0, 1, 2]},
            ),
            "quantum_efficiency": xr.DataArray(
                [0.5, 0.8, 0.9],
                dims="quantum_efficiency_id",
                coords={"quantum_efficiency_id": [0, 1, 2]},
            ),
        }
    )
    xr.testing.assert_equal(parameters, exp_parameters)

    # Check '.logs'
    logs = result.logs
    assert isinstance(logs, xr.Dataset)

    exp_logs = xr.Dataset(
        {
            "level": xr.DataArray(
                ([100] * 3 + [200] * 3 + [300] * 3) * 3,
                dims="id",
            ),
            "quantum_efficiency": xr.DataArray(
                [0.5, 0.8, 0.9] * 9,
                dims="id",
            ),
        },
        coords={"id": range(27)},
    )
    xr.testing.assert_equal(logs, exp_logs)


@pytest.mark.parametrize("with_dask", [False, True])
def test_product(with_dask: bool):
    """Test running mode 'observation'."""
    filename = Path("tests/running_mode/data/observation_product.yaml")
    assert filename.exists()

    cfg = pyxel.load(filename)
    assert isinstance(cfg, Configuration)

    observation = cfg.observation
    detector = cfg.detector
    pipeline = cfg.pipeline

    observation.with_dask = with_dask

    assert isinstance(observation, Observation)
    assert observation.parameter_mode == ParameterMode.Product

    assert isinstance(detector, CCD)
    assert isinstance(pipeline, DetectionPipeline)

    result = pyxel.observation_mode(
        observation=observation, detector=detector, pipeline=pipeline
    )
    assert isinstance(result, ObservationResult)

    # Check '.dataset'
    ds = result.dataset
    assert isinstance(ds, xr.Dataset)

    assert list(ds["image"].coords) == [
        "readout_time",
        "y",
        "x",
        "illumination_uniform.level",
        "illumination_elliptic.level",
        "quantum_efficiency",
    ]
    assert list(ds["image"].coords["readout_time"].values) == [1]
    assert list(ds["image"].coords["y"].values) == list(range(100))
    assert list(ds["image"].coords["x"].values) == list(range(100))
    assert list(ds["image"].coords["illumination_uniform.level"].values) == [10, 20, 30]
    assert list(ds["image"].coords["illumination_elliptic.level"].values) == [
        100,
        200,
        300,
    ]
    assert list(ds["image"].coords["quantum_efficiency"].values) == [0.5, 0.8, 0.9]

    # Check '.parameters'
    parameters = result.parameters
    assert isinstance(parameters, xr.Dataset)

    exp_parameters = xr.Dataset(
        {
            "illumination_uniform.level": xr.DataArray(
                [10, 20, 30],
                dims="illumination_uniform.level_id",
                coords={"illumination_uniform.level_id": [0, 1, 2]},
            ),
            "illumination_elliptic.level": xr.DataArray(
                [100, 200, 300],
                dims="illumination_elliptic.level_id",
                coords={"illumination_elliptic.level_id": [0, 1, 2]},
            ),
            "quantum_efficiency": xr.DataArray(
                [0.5, 0.8, 0.9],
                dims="quantum_efficiency_id",
                coords={"quantum_efficiency_id": [0, 1, 2]},
            ),
        }
    )
    xr.testing.assert_equal(parameters, exp_parameters)

    # Check '.logs'
    logs = result.logs
    assert isinstance(logs, xr.Dataset)

    exp_logs = xr.Dataset(
        {
            "level": xr.DataArray(
                ([100] * 3 + [200] * 3 + [300] * 3) * 3,
                dims="id",
            ),
            "quantum_efficiency": xr.DataArray(
                [0.5, 0.8, 0.9] * 9,
                dims="id",
            ),
        },
        coords={"id": range(27)},
    )
    xr.testing.assert_equal(logs, exp_logs)


@pytest.mark.deprecated
@pytest.mark.parametrize("with_dask", [False, True])
def test_sequential_simple_deprecated(with_dask: bool):
    """Test running mode 'sequential'."""
    filename = Path(
        "tests/running_mode/data/deprecated_observation_sequential_simple.yaml"
    )
    assert filename.exists()

    cfg = pyxel.load(filename)
    assert isinstance(cfg, Configuration)

    observation = cfg.observation
    detector = cfg.detector
    pipeline = cfg.pipeline

    observation.with_dask = with_dask

    assert isinstance(observation, Observation)
    assert observation.parameter_mode == ParameterMode.Sequential

    assert isinstance(detector, CCD)
    assert isinstance(pipeline, DetectionPipeline)

    result = pyxel.observation_mode(
        observation=observation, detector=detector, pipeline=pipeline
    )
    assert isinstance(result, ObservationResult)

    # Check '.dataset'
    ds_dict = result.dataset
    assert isinstance(ds_dict, dict)

    assert list(ds_dict) == ["level", "quantum_efficiency"]

    assert isinstance(ds_dict["level"], xr.Dataset)

    assert list(ds_dict["level"]["image"].coords) == ["readout_time", "y", "x", "level"]

    assert list(ds_dict["level"]["image"].coords["readout_time"].values) == [1]
    assert list(ds_dict["level"]["image"].coords["y"].values) == list(range(100))
    assert list(ds_dict["level"]["image"].coords["x"].values) == list(range(100))
    assert list(ds_dict["level"]["image"].coords["level"].values) == [10, 20, 30]

    assert list(ds_dict["quantum_efficiency"]["image"].coords) == [
        "readout_time",
        "y",
        "x",
        "quantum_efficiency",
    ]

    assert list(
        ds_dict["quantum_efficiency"]["image"].coords["readout_time"].values
    ) == [1]
    assert list(ds_dict["quantum_efficiency"]["image"].coords["y"].values) == list(
        range(100)
    )
    assert list(ds_dict["quantum_efficiency"]["image"].coords["x"].values) == list(
        range(100)
    )
    assert list(
        ds_dict["quantum_efficiency"]["image"].coords["quantum_efficiency"].values
    ) == [0.5, 0.8, 0.9]

    # Check '.parameters'
    parameters = result.parameters
    assert isinstance(parameters, xr.Dataset)

    exp_parameters = xr.Dataset(
        {
            "level": xr.DataArray(
                [10, 20, 30], dims="level_id", coords={"level_id": [0, 1, 2]}
            ),
            "quantum_efficiency": xr.DataArray(
                [0.5, 0.8, 0.9],
                dims="quantum_efficiency_id",
                coords={"quantum_efficiency_id": [0, 1, 2]},
            ),
        }
    )
    xr.testing.assert_equal(parameters, exp_parameters)

    # Check '.logs'
    logs = result.logs
    assert isinstance(logs, xr.Dataset)

    exp_logs = xr.Dataset(
        {
            "level": xr.DataArray(
                [10.0, 20.0, 30.0, np.nan, np.nan, np.nan], dims="id"
            ),
            "quantum_efficiency": xr.DataArray(
                [np.nan, np.nan, np.nan, 0.5, 0.8, 0.9], dims="id"
            ),
        },
        coords={"id": [0, 1, 2, 3, 4, 5]},
    )
    xr.testing.assert_equal(logs, exp_logs)


@pytest.mark.parametrize("with_dask", [False, True])
def test_sequential_simple(with_dask: bool):
    """Test running mode 'sequential'."""
    filename = Path("tests/running_mode/data/observation_sequential_simple.yaml")
    assert filename.exists()

    cfg = pyxel.load(filename)
    assert isinstance(cfg, Configuration)

    observation = cfg.observation
    detector = cfg.detector
    pipeline = cfg.pipeline

    observation.with_dask = with_dask

    assert isinstance(observation, Observation)
    assert observation.parameter_mode == ParameterMode.Sequential

    assert isinstance(detector, CCD)
    assert isinstance(pipeline, DetectionPipeline)

    result = pyxel.observation_mode(
        observation=observation, detector=detector, pipeline=pipeline
    )
    assert isinstance(result, ObservationResult)

    # Check '.dataset'
    ds_dict = result.dataset
    assert isinstance(ds_dict, dict)

    assert list(ds_dict) == ["level", "quantum_efficiency"]

    assert isinstance(ds_dict["level"], xr.Dataset)

    assert list(ds_dict["level"]["image"].coords) == ["readout_time", "y", "x", "level"]

    assert list(ds_dict["level"]["image"].coords["readout_time"].values) == [1]
    assert list(ds_dict["level"]["image"].coords["y"].values) == list(range(100))
    assert list(ds_dict["level"]["image"].coords["x"].values) == list(range(100))
    assert list(ds_dict["level"]["image"].coords["level"].values) == [10, 20, 30]

    assert list(ds_dict["quantum_efficiency"]["image"].coords) == [
        "readout_time",
        "y",
        "x",
        "quantum_efficiency",
    ]

    assert list(
        ds_dict["quantum_efficiency"]["image"].coords["readout_time"].values
    ) == [1]
    assert list(ds_dict["quantum_efficiency"]["image"].coords["y"].values) == list(
        range(100)
    )
    assert list(ds_dict["quantum_efficiency"]["image"].coords["x"].values) == list(
        range(100)
    )
    assert list(
        ds_dict["quantum_efficiency"]["image"].coords["quantum_efficiency"].values
    ) == [0.5, 0.8, 0.9]

    # Check '.parameters'
    parameters = result.parameters
    assert isinstance(parameters, xr.Dataset)

    exp_parameters = xr.Dataset(
        {
            "level": xr.DataArray(
                [10, 20, 30], dims="level_id", coords={"level_id": [0, 1, 2]}
            ),
            "quantum_efficiency": xr.DataArray(
                [0.5, 0.8, 0.9],
                dims="quantum_efficiency_id",
                coords={"quantum_efficiency_id": [0, 1, 2]},
            ),
        }
    )
    xr.testing.assert_equal(parameters, exp_parameters)

    # Check '.logs'
    logs = result.logs
    assert isinstance(logs, xr.Dataset)

    exp_logs = xr.Dataset(
        {
            "level": xr.DataArray(
                [10.0, 20.0, 30.0, np.nan, np.nan, np.nan], dims="id"
            ),
            "quantum_efficiency": xr.DataArray(
                [np.nan, np.nan, np.nan, 0.5, 0.8, 0.9], dims="id"
            ),
        },
        coords={"id": [0, 1, 2, 3, 4, 5]},
    )
    xr.testing.assert_equal(logs, exp_logs)


@pytest.mark.skip(reason="Fix this test")
@pytest.mark.parametrize("with_dask", [False, True])
def test_sequential(with_dask: bool):
    """Test running mode 'sequential'."""
    filename = Path("tests/running_mode/data/observation_sequential.yaml")
    assert filename.exists()

    cfg = pyxel.load(filename)
    assert isinstance(cfg, Configuration)

    observation = cfg.observation
    detector = cfg.detector
    pipeline = cfg.pipeline

    observation.with_dask = with_dask

    assert isinstance(observation, Observation)
    assert observation.parameter_mode == ParameterMode.Sequential

    assert isinstance(detector, CCD)
    assert isinstance(pipeline, DetectionPipeline)

    result = pyxel.observation_mode(
        observation=observation, detector=detector, pipeline=pipeline
    )
    assert isinstance(result, ObservationResult)

    # Check '.dataset'
    ds_dict = result.dataset
    assert isinstance(ds_dict, dict)

    assert list(ds_dict) == [
        "illumination_uniform.level",
        "illumination_elliptic.level",
        "quantum_efficiency",
    ]

    assert isinstance(ds_dict["illumination_uniform"], xr.Dataset)

    assert list(ds_dict["illumination_uniform.level"]["image"].coords) == [
        "readout_time",
        "y",
        "x",
        "illumination_uniform.level",
    ]

    assert list(
        ds_dict["illumination_uniform.level"]["image"].coords["readout_time"].values
    ) == [1]
    assert list(
        ds_dict["illumination_uniform.level"]["image"].coords["y"].values
    ) == list(range(100))
    assert list(
        ds_dict["illumination_uniform.level"]["image"].coords["x"].values
    ) == list(range(100))
    assert list(
        ds_dict["illumination_uniform.level"]["image"]
        .coords["illumination_uniform.level"]
        .values
    ) == [10, 20, 30]

    assert list(ds_dict["quantum_efficiency"]["image"].coords) == [
        "readout_time",
        "y",
        "x",
        "quantum_efficiency",
    ]

    assert list(
        ds_dict["quantum_efficiency"]["image"].coords["readout_time"].values
    ) == [1]
    assert list(ds_dict["quantum_efficiency"]["image"].coords["y"].values) == list(
        range(100)
    )
    assert list(ds_dict["quantum_efficiency"]["image"].coords["x"].values) == list(
        range(100)
    )
    assert list(
        ds_dict["quantum_efficiency"]["image"].coords["quantum_efficiency"].values
    ) == [0.5, 0.8, 0.9]

    # Check '.parameters'
    parameters = result.parameters
    assert isinstance(parameters, xr.Dataset)

    exp_parameters = xr.Dataset(
        {
            "level": xr.DataArray(
                [10, 20, 30], dims="level_id", coords={"level_id": [0, 1, 2]}
            ),
            "quantum_efficiency": xr.DataArray(
                [0.5, 0.8, 0.9],
                dims="quantum_efficiency_id",
                coords={"quantum_efficiency_id": [0, 1, 2]},
            ),
        }
    )
    xr.testing.assert_equal(parameters, exp_parameters)

    # Check '.logs'
    logs = result.logs
    assert isinstance(logs, xr.Dataset)

    exp_logs = xr.Dataset(
        {
            "level": xr.DataArray(
                [10.0, 20.0, 30.0, np.nan, np.nan, np.nan], dims="id"
            ),
            "quantum_efficiency": xr.DataArray(
                [np.nan, np.nan, np.nan, 0.5, 0.8, 0.9], dims="id"
            ),
        },
        coords={"id": [0, 1, 2, 3, 4, 5]},
    )
    xr.testing.assert_equal(logs, exp_logs)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Charge Distortion Model for :term:`CCDs<CCD>`.

============================
This is a function to run the upgraded :term:`CDM` :term:`CTI` model developed by Alex Short (ESA).

:requires: NumPy

:author: Alex Short
:author: Sami-Matias Niemi
:author: David Lucsanyi

Descriptions of model quantities in the paper:

Ne - number of electrons in a pixel
ne - electron density in the vicinity of the trap
Vc - volume of the charge cloud
Pr - the probability that the trap will release the electron into the sample
tau_c - capture time constant
Pc - capture probability (per vacant trap) as a function of the number of sample electrons Ne
NT - number of traps in the column,
NT = 2*nt*Vg*x  where x is the number of TDI transfers or the column length in pixels.
Nc - number of electrons captured by a given trap species during the transit of an integrating signal packet
N0 - initial trap occupancy
Nr - number of electrons released into the sample during a transit along the column
vg: assumed maximum geometrical volume electrons can occupy within a pixel (parallel)
svg: assumed maximum geometrical volume electrons can occupy within a pixel (serial)
t: constant TDI period (parallel)
st: constant TDI period (serial)
"""

from collections.abc import Sequence
from enum import Enum
from typing import Literal, Optional

import astropy.constants as const
import numba
import numpy as np

from pyxel.detectors import CCD


class CDMdirection(Enum):
    """:term:`CDM` direction class."""

    Parallel = "parallel"
    Serial = "serial"


def cdm(
    detector: CCD,
    direction: Literal["parallel", "serial"],
    beta: float,
    trap_release_times: Sequence[float],
    trap_densities: Sequence[float],
    sigma: Sequence[float],
    full_well_capacity: Optional[float] = None,
    max_electron_volume: float = 0.0,
    transfer_period: float = 0.0,
    charge_injection: bool = False,
    electron_effective_mass: float = 0.5,
) -> None:
    r"""Charge Distortion Model (CDM) model wrapper.

    Parameters
    ----------
    detector : CCD
        Pyxel CCD detector object.
    direction : literal
        Set ``"parallel"`` for :term:`CTI` in parallel direction or ``"serial"`` for :term:`CTI` in serial register.
    beta : float
        Electron cloud expansion coefficient :math:`\beta`.
    trap_release_times : sequence of float
        Trap release time constants :math:`\tau_r`. Unit: :math:`s`.
    trap_densities : sequence of float
        Absolute trap densities :math:`n_t`. Unit: :math:`cm^{-3}`.
    sigma : sequence of float
        Trap capture cross section :math:`\sigma`. Unit: :math:`cm^2`.
    full_well_capacity : float
        Full well capacity :math:`FWC`. Unit: :math:`e^-`.
    max_electron_volume : float
        Maximum geometrical volume :math:`V_g` that electrons can occupy within a pixel. Unit: :math:`cm^3`.
    transfer_period : float
        Transfer period :math:`t` (TDI period). Unit: :math:`s`.
    charge_injection : bool
        Enable charge injection (only used in ``"parallel"`` mode).
    electron_effective_mass : float
        Electron effective mass in the semiconductor lattice. Unit: 1 electron mass
    """

    e_effective_mass = electron_effective_mass * const.m_e.value

    if full_well_capacity is None:
        fwc_final = detector.characteristics.full_well_capacity
    else:
        fwc_final = full_well_capacity

    mode = CDMdirection(direction)

    if not isinstance(detector, CCD):
        # Later, this will be checked in when YAML configuration file is parsed
        raise TypeError("Expecting a `CCD` object for 'detector'.")

    if not (0.0 <= max_electron_volume <= 1.0):
        raise ValueError("'max_electron_volume' must be between 0.0 and 1.0.")
    if not (0.0 <= beta <= 1.0):
        raise ValueError("'beta' must be between 0.0 and 1.0.")
    if not (0.0 <= fwc_final <= 1.0e7):
        raise ValueError("'full_well_capacity' must be between 0 and 1e7.")
    if not (0.0 <= transfer_period <= 10.0):
        raise ValueError("'transfer_period' must be between 0.0 and 10.0.")
    if not (len(trap_densities) == len(trap_release_times) == len(sigma)):
        raise ValueError(
            "Length of 'sigma', 'trap_densities' and 'trap_release_times' not the same!"
        )
    if len(trap_release_times) == 0:
        raise ValueError("Expecting inputs for at least one trap species.")

    e_thermal_velocity = np.sqrt(
        3 * const.k_B.value * detector.environment.temperature / e_effective_mass
    )

    if mode == CDMdirection.Parallel:
        detector.pixel.array = run_cdm_parallel(
            array=detector.pixel.array,
            vg=max_electron_volume,
            t=transfer_period,
            fwc=fwc_final,
            vth=e_thermal_velocity,
            charge_injection=charge_injection,
            chg_inj_parallel_transfers=detector.geometry.row,
            beta=beta,
            tr=np.array(trap_release_times),
            nt=np.array(trap_densities),
            sigma=np.array(sigma),
        )

    elif mode == CDMdirection.Serial:
        detector.pixel.array = run_cdm_serial(
            array=detector.pixel.array,
            vg=max_electron_volume,
            t=transfer_period,
            fwc=fwc_final,
            vth=e_thermal_velocity,
            beta=beta,
            tr=np.array(trap_release_times),
            nt=np.array(trap_densities),
            sigma=np.array(sigma),
        )


@numba.njit(nogil=True)
def run_cdm_parallel(
    array: np.ndarray,
    beta: float,
    vg: float,
    t: float,
    fwc: float,
    vth: float,
    tr: np.ndarray,
    nt: np.ndarray,
    sigma: np.ndarray,
    charge_injection: bool = False,
    chg_inj_parallel_transfers: int = 0,
) -> np.ndarray:
    r"""Run :term:`CDM` in parallel direction.

    Parameters
    ----------
    array: ndarray
        Input array.
    beta: float
        Electron cloud expansion coefficient :math:`\beta`.
    vg: float
        Maximum geometrical volume :math:`V_g` that electrons can occupy within a pixel. Unit: :math:`cm^3`.
    t: float
        Transfer period :math:`t` (TDI period). Unit: :math:`s`.
    fwc: float
        Full well capacity :math:`FWC`. Unit: :math:`e^-`.
    vth: float
        Electron thermal velocity.
    tr: sequence of float
        Trap release time constants :math:`\tau_r`. Unit: :math:`s`.
    nt: sequence of float
        Absolute trap densities :math:`n_t`. Unit: :math:`cm^{-3}`.
    sigma: sequence of float
        Trap capture cross section :math:`\sigma`. Unit: :math:`cm^2`.
    charge_injection: bool
        Enable charge injection (only used in "parallel" mode).
    chg_inj_parallel_transfers: int
        Number of parallel transfers for charge injection.

    Returns
    -------
    array: ndarray
        Output array.
    """
    ydim, xdim = array.shape  # full signal array we want to apply cdm for
    kdim_p = len(nt)
    # np.clip(s, 0., fwc, s)      # full well capacity
    nt = nt / vg  # parallel trap density (traps / cm**3)
    # nt_p *= rdose             # absolute trap density [per cm**3]

    # IMAGING (non-TDI) MODE
    # Parallel direction
    no = np.zeros((xdim, kdim_p))
    alpha_p: np.ndarray = t * sigma * vth * fwc**beta / (2.0 * vg)
    g_p: np.ndarray = 2.0 * nt * vg / fwc**beta
    for i in range(0, ydim):
        if charge_injection:
            gamma_p = (
                g_p * chg_inj_parallel_transfers
            )  # number of all transfers in parallel dir.
        else:
            gamma_p = g_p * i  # TODO: (i+1) ?????????????
        for k in range(kdim_p):
            for j in range(xdim):
                nc = 0.0
                if array[i, j] > 0.01:
                    nc = max(
                        (gamma_p[k] * array[i, j] ** beta - no[j, k])
                        / (gamma_p[k] * array[i, j] ** (beta - 1.0) + 1.0)
                        * (1.0 - np.exp(-1 * alpha_p[k] * array[i, j] ** (1.0 - beta))),
                        0.0,
                    )
                    no[j, k] += nc
                nr = no[j, k] * (1.0 - np.exp(-t / tr[k]))
                array[i, j] += -1 * nc + nr
                no[j, k] -= nr
                if array[i, j] < 0.01:
                    array[i, j] = 0.0

    return array


@numba.njit(nogil=True)
def run_cdm_serial(
    array: np.ndarray,
    beta: float,
    vg: float,
    t: float,
    fwc: float,
    vth: float,
    tr: np.ndarray,
    nt: np.ndarray,
    sigma: np.ndarray,
) -> np.ndarray:
    r"""Run :term:`CDM` in serial direction.

    Parameters
    ----------
    array: ndarray
        Input array.
    beta: float
        Electron cloud expansion coefficient :math:`\beta`.
    vg: float
        Maximum geometrical volume :math:`V_g` that electrons can occupy within a pixel. Unit: :math:`cm^3`.
    t: float
        Transfer period :math:`t` (TDI period). Unit: :math:`s`.
    fwc: float
        Full well capacity :math:`FWC`. Unit: :math:`e^-`.
    vth: float
        Electron thermal velocity.
    tr: sequence of float
        Trap release time constants :math:`\tau_r`. Unit: :math:`s`.
    nt: sequence of float
        Absolute trap densities :math:`n_t`. Unit: :math:`cm^{-3}`.
    sigma: sequence of float
        Trap capture cross section :math:`\sigma`. Unit: :math:`cm^2`.

    Returns
    -------
    array: ndarray
    """
    ydim, xdim = array.shape  # full signal array we want to apply cdm for
    kdim_s = len(nt)
    # np.clip(s, 0., fwc, s)      # full well capacity
    nt = nt / vg  # serial trap density (traps / cm**3)
    # nt_s *= rdose             # absolute trap density [per cm**3]

    # IMAGING (non-TDI) MODE
    # Serial direction

    sno = np.zeros((ydim, kdim_s))
    alpha_s: np.ndarray = t * sigma * vth * fwc**beta / (2.0 * vg)
    g_s = 2.0 * nt * vg / fwc**beta
    for j in range(0, xdim):
        gamma_s = g_s * j  # TODO: (j+1) ?????????????
        for k in range(kdim_s):
            for i in range(ydim):
                nc = 0.0
                if array[i, j] > 0.01:
                    nc = max(
                        (gamma_s[k] * array[i, j] ** beta - sno[i, k])
                        / (gamma_s[k] * array[i, j] ** (beta - 1.0) + 1.0)
                        * (1.0 - np.exp(-1 * alpha_s[k] * array[i, j] ** (1.0 - beta))),
                        0.0,
                    )
                    sno[i, k] += nc
                nr = sno[i, k] * (1.0 - np.exp(-t / tr[k]))
                array[i, j] += -1 * nc + nr
                sno[i, k] -= nr
                if array[i, j] < 0.01:
                    array[i, j] = 0.0

    return array


#
# def plot_serial_profile(
#     data: np.ndarray, row: int, data2: Optional[np.ndarray] = None
# ) -> None:
#     """TBW.
#
#     :param data:
#     :param row:
#     :param data2:
#     """
#     ydim, xdim = data.shape
#     profile_x = list(range(ydim))
#     profile_y = data[row, :]
#     plt.title("Serial profile")
#     plt.plot(profile_x, profile_y, color="blue")
#     if data2 is not None:
#         profile_y_2 = data2[row, :]
#         plt.plot(profile_x, profile_y_2, color="red")
#
#
# def plot_parallel_profile(
#     data: np.ndarray, col: int, data2: Optional[np.ndarray] = None
# ) -> None:
#     """TBW.
#
#     :param data:
#     :param col:
#     :param data2:
#     """
#     ydim, xdim = data.shape
#     profile_x = list(range(xdim))
#     profile_y = data[:, col]
#     plt.title("Parallel profile")
#     plt.plot(profile_x, profile_y, color="blue")
#     if data2 is not None:
#         profile_y_2 = data2[:, col]
#         plt.plot(profile_x, profile_y_2, color="red")
#
#
# def plot_1d_profile(
#     array: np.ndarray, offset: int = 0, label: str = "", m: str = "-"
# ) -> None:
#     """Plot profile on log scale.
#
#     :param array:
#     :param offset:
#     :param label:
#     :param m:
#     """
#     x = list(range(offset, offset + len(array)))
#     # plt.title('Parallel profile, charge injection')
#     plt.semilogy(x, array, m, label=label)
#     if label:
#         plt.legend()
#
#
# def plot_1d_profile_lin(
#     array: np.ndarray,
#     offset: int = 0,
#     label: str = "",
#     m: str = "-",
#     col: Optional[str] = None,
# ) -> None:
#     """TBW.
#
#     :param array:
#     :param offset:
#     :param label:
#     :param m:
#     :param col:
#     """
#     x = list(range(offset, offset + len(array)))
#     # plt.title('Parallel profile, charge injection')
#     plt.plot(x, array, m, label=label, color=col)
#     if label:
#         plt.legend()
#
#
# def plot_1d_profile_with_err(
#     array: np.ndarray, error: np.ndarray, offset: int = 0, label: str = ""
# ) -> None:
#     """TBW.
#
#     :param array:
#     :param error:
#     :param offset:
#     :param label:
#     """
#     x = list(range(offset, offset + len(array)))
#     plt.title("Parallel profile with error, charge injection")
#     plt.errorbar(x, array, error, label=label, linestyle="None", marker=".")
#     if label:
#         plt.legend()
#
#
# def plot_residuals(
#     data: np.ndarray, data2: np.ndarray, label: str = ""
# ) -> None:  # col='magenta',
#     """TBW.
#
#     :param data:
#     :param data2:
#     # :param col:
#     :param label:
#     """
#     x = list(range(len(data)))
#     # plt.title('Residuals of fitted and target parallel CTI profiles')
#     residuals = np.around(data - data2, decimals=5)
#     # residuals = data-data2
#     # plt.plot(x, residuals, '.', color=col, label=label)
#     plt.plot(x, residuals, ".", label=label)
#     plt.legend()
#
#
# def plot_image(data: np.ndarray) -> None:
#     """TBW.
#
#     :param data:
#     """
#     plt.imshow(data, cmap=plt.gray())  # , interpolation='nearest')
#     plt.xlabel("x - serial direction")
#     plt.ylabel("y - parallel direction")
#     plt.title("CCD image with CTI")
#     plt.colorbar()


# # @numba.jit(nopython=True, nogil=True, parallel=True)
# def optimized_cdm(
#     s: np.ndarray,
#     beta_p: float,
#     beta_s: float,
#     vg: float,
#     svg: float,
#     t: float,
#     st: float,
#     fwc: float,
#     sfwc: float,
#     vth: float,
#     tr_p: np.ndarray,
#     tr_s: np.ndarray,
#     nt_p: np.ndarray,
#     nt_s: np.ndarray,
#     sigma_p: np.ndarray,
#     sigma_s: np.ndarray,
#     charge_injection: bool = False,
#     chg_inj_parallel_transfers: int = 0,
#     parallel_cti: bool = True,
#     serial_cti: bool = True,
# ) -> np.ndarray:
#     """CDM model.
#
#     Done by Patricia Liebing. Not yet test or compared to the original 'run_cdm' function.
#
#     :param s: np.ndarray
#     :param dob:
#     :param beta_p: electron cloud expansion coefficient (parallel)
#     :param beta_s: electron cloud expansion coefficient (serial)
#     :param vg: assumed maximum geometrical volume electrons can occupy within
#                a pixel (parallel)
#     :param svg: assumed maximum geometrical volume electrons can occupy within
#                 a pixel (serial)
#     :param t: constant TDI period (parallel)
#     :param st: constant TDI period (serial)
#     :param fwc:
#     :param sfwc:
#     :param vth:
#     :param charge_injection:
#     :param chg_inj_parallel_transfers:
#     :param sigma_p:
#     :param sigma_s:
#     :param tr_p:
#     :param tr_s:
#     :param nt_p: number of traps per electron cloud (and not pixel!)
#                  in parallel direction
#     :param nt_s: number of traps per electron cloud (and not pixel!)
#                  in serial direction
#     :param parallel_cti:
#     :param serial_cti:
#     :return:
#     """
#     ydim, xdim = s.shape  # full signal array we want to apply cdm for
#
#     kdim_p = len(nt_p)
#     kdim_s = len(nt_s)
#
#     # np.clip(s, 0., fwc, s)      # full well capacity
#
#     nt_p = nt_p / vg  # parallel trap density (traps / cm**3)
#     nt_s = nt_s / svg  # serial trap density (traps / cm**3)
#
#     # nt_p *= rdose             # absolute trap density [per cm**3]
#     # nt_s *= rdose             # absolute trap density [per cm**3]
#
#     # IMAGING (non-TDI) MODE
#     # Parallel direction
#     if parallel_cti:
#         no = np.zeros((xdim, kdim_p))
#         alpha_p = t * sigma_p * vth * fwc ** beta_p / (2.0 * vg)  # type: np.ndarray
#         g_p = 2.0 * nt_p * vg / fwc ** beta_p  # type: np.ndarray
#         for i in range(0, ydim):
#             if charge_injection:
#                 gamma_p = (
#                     g_p * chg_inj_parallel_transfers
#                 )  # number of all transfers in parallel dir.
#             else:
#                 gamma_p = g_p * (i + 1)
#             for k in range(kdim_p):
#                 nc = np.zeros(xdim)
#                 nctest = np.zeros(xdim)
#                 sclm = s[i, :]
#                 idx = np.where(sclm > 0.01)
#                 nok = no[:, k]
#                 nc[idx] += (
#                     (gamma_p[k] * sclm[idx] ** beta_p - nok[idx])
#                     / (gamma_p[k] * sclm[idx] ** (beta_p - 1.0) + 1.0)
#                     * (1.0 - np.exp(-1 * alpha_p[k] * sclm[idx] ** (1.0 - beta_p)))
#                 )
#                 no[:, k] += np.fmax(nctest, nc)
#                 nr = no[:, k] * (1.0 - np.exp(-t / tr_p[k]))
#                 s[i, :] += -1 * nc + nr
#                 no[:, k] -= nr
#
#     # IMAGING (non-TDI) MODE
#     # Serial direction
#     if serial_cti:
#         sno = np.zeros((ydim, kdim_s))
#         alpha_s = st * sigma_s * vth * sfwc ** beta_s / (2.0 * svg)  # type: np.ndarray
#         g_s = 2.0 * nt_s * svg / sfwc ** beta_s  # type: np.ndarray
#         for j in range(0, xdim):
#             gamma_s = g_s * (j + 1)
#             for k in range(kdim_s):
#                 ncs = np.zeros(ydim)
#                 ncstest = np.zeros(ydim)
#                 srow = s[:, j]
#                 idxs = np.where(srow > 0.01)
#                 snok = sno[:, k]
#                 ncs[idxs] += (
#                     (gamma_s[k] * srow[idxs] ** beta_s - snok[idxs])
#                     / (gamma_s[k] * srow[idxs] ** (beta_s - 1.0) + 1.0)
#                     * (1.0 - np.exp(-1 * alpha_s[k] * srow[idxs] ** (1.0 - beta_s)))
#                 )
#                 sno[:, k] += np.fmax(ncstest, ncs)
#                 nrs = sno[:, k] * (1.0 - np.exp(-st / tr_s[k]))
#                 s[:, j] += -1 * ncs + nrs
#                 sno[:, k] -= nrs
#
#     return s



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


import sys

import pytest

from pyxel.util import memory_usage_details


def test_no_attributes(ccd_empty):
    with pytest.raises(ValueError, match=r"No attributes provided"):
        memory_usage_details(ccd_empty)


def test_invalid_attribute(ccd_empty):
    with pytest.raises(AttributeError):
        memory_usage_details(ccd_empty, ["dummy"])


@pytest.mark.skip(reason="Fix this test !")
@pytest.mark.skipif(
    sys.version_info < (3, 8), reason="Different value for older versions."
)
def test_memory_usage(ccd_empty):
    attributes = [
        "_photon",
        "_charge",
        "_pixel",
        "_signal",
        "_image",
        "environment",
        "_geometry",
        "_characteristics",
    ]

    usage = memory_usage_details(ccd_empty, attributes, print_result=False)

    empty_size = {
        "characteristics": 1160,
        "charge": 43896,
        "environment": 512,
        "geometry": 664,
        "image": 432,
        "pixel": 432,
        "signal": 432,
    }

    assert usage == empty_size



#   Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#  #
#   This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#   is part of this Pyxel package. No part of the package, including
#   this file, may be copied, modified, propagated, or distributed except according to
#   the terms contained in the file ‘LICENCE.txt’.

"""Simple models to simulate charge deposition by ionizing particles (e.g. cosmic rays)."""

from collections.abc import Sequence
from pathlib import Path
from typing import Literal, Optional, Union

import numpy as np
import pandas as pd

from pyxel.data_structure import Charge
from pyxel.detectors import Detector
from pyxel.util import materials, set_random_seed


def charge_deposition(
    detector: Detector,
    flux: float,
    step_size: float = 1.0,
    energy_mean: float = 1.0,
    energy_spread: float = 0.1,
    energy_spectrum: Union[str, Path, None] = None,
    energy_spectrum_sampling: Literal["linear", "log"] = "log",
    ehpair_creation: float = 3.65,
    material_density: float = 2.3290,
    particle_direction: Literal["isotropic", "orthogonal"] = "isotropic",
    stopping_power_curve: Union[str, Path, None] = None,
    seed: Optional[int] = None,
) -> None:
    """Simulate charge deposition by ionizing particles using a stopping power curve.

    Parameters
    ----------
    detector : Detector
        the detector
    flux : float
        the flux of incoming particles in particle/s
    step_size : float
        the size of the considered unitary step in unit length along which energy is deposited
    energy_mean : float
        the mean energy of the incoming ionizing particles in MeV
    energy_spread : float
        the spread in energy of the incoming ionizing particles in MeV
    energy_spectrum : str
        the location of the file describing the energy spectrum of incident particles
        if no spectrum is provided energies are randomly drawn from a normal distribution
        with mean and spread defined above
        note that the energy spectrum is assumed to be a txt file with two columns [energy, flux]
        with the energy in MeV
    energy_spectrum_sampling : str
        "log" or None: the energy spectrum is sampled in log space
        "linear" : the energy spectrum is sampled in linear space
    ehpair_creation : float
        the energy required to generate a electron-hole pair in eV
        by default the Si value at room temperature is parsed i.e. 3.65 eV
    material_density : float
        the material density in g/cm3
        by default he Si value at room temperature is parsed i.e. 2.3290 g/cm3
    particle_direction : str
        "isotropic" : particles are coming from all directions (outside of the sensor)
        "orthogonal" : particles are coming from the top of the sensor (thickness = 0) and orthogonal to its surface
    stopping_power_curve : str
        the location of the file describing the total massive stopping power
        energetic loss per mass of material and per unit path length versus particle energy
        note that the the stopping power curve is assumed to be a csv file with two columns [energy, stopping power]
        energy in MeV, stopping power in MeV cm2/g
    seed : int, optional
    """
    with set_random_seed(seed):
        tracks = simulate_charge_deposition(
            flux=flux,
            exposure=detector.time_step,
            x_lim=detector.geometry.horz_dimension,
            y_lim=detector.geometry.vert_dimension,
            z_lim=detector.geometry.total_thickness,
            step_size=step_size,
            energy_mean=energy_mean,
            energy_spread=energy_spread,
            energy_spectrum=energy_spectrum,
            energy_spectrum_sampling=energy_spectrum_sampling,
            ehpair_creation=ehpair_creation,
            material_density=material_density,
            particle_direction=particle_direction,
            stopping_power_curve=stopping_power_curve,
        )

    detector.charge.add_charge_dataframe(tracks_to_charge(tracks))


def charge_deposition_in_mct(
    detector: Detector,
    flux: float,
    step_size: float = 1.0,
    energy_mean: float = 1.0,
    energy_spread: float = 0.1,
    energy_spectrum: Union[str, Path, None] = None,
    energy_spectrum_sampling: Literal["linear", "log"] = "log",
    cutoff_wavelength: float = 2.5,
    particle_direction: Literal["isotropic", "orthogonal"] = "isotropic",
    stopping_power_curve: Union[str, Path, None] = None,
    seed: Optional[int] = None,
) -> None:
    """Simulate charge deposition by ionizing particles using a stopping power curve.

    Parameters
    ----------
    detector : Detector
        the detector
    flux : float
        the flux of incoming particles in particle/s
    step_size : float
        the size of the considered unitary step in unit length along which energy is deposited
    energy_mean : float
        the mean energy of the incoming ionizing particles in MeV
    energy_spread : float
        the spread in energy of the incoming ionizing particles in MeV
    energy_spectrum : str
        the location of the file describing the energy spectrum of incident particles
        if no spectrum is provided energies are randomly drawn from a normal distribution
        with mean and spread defined above
        note that the energy spectrum is assumed to be a txt file with two columns [energy, flux]
        with the energy in MeV
    energy_spectrum_sampling : str. Default: 'log'
        "log" : the energy spectrum is sampled in log space
        "linear" : the energy spectrum is sampled in linear space
    cutoff_wavelength : float
        the longest wavelength in micrometer at which the QE reaches 50% of its maximum,
        used to compute the bandgap energy, and the corresponding fraction of cadmium
    particle_direction : str. Default: 'isotropic'
        "isotropic" : particles are coming from all directions (outside of the sensor)
        "orthogonal" : particles are coming from the top of the sensor (thickness = 0) and orthogonal to its surface
    stopping_power_curve : str
        the location of the file describing the total massive stopping power
        energetic loss per mass of material and per unit path length versus particle energy
        note that the the stopping power curve is assumed to be a csv file with two columns [energy, stopping power]
        energy in MeV, stopping power in MeV cm2/g
    seed : int, optional
    """
    lambdae = materials.lambda_e(cutoff_wavelength)
    eg = 1.24 / lambdae
    ehpair_creation = 3 * eg
    x = materials.eg_hansen_inverse(eg, detector.environment.temperature)
    material_density = materials.density(x)

    with set_random_seed(seed):
        tracks = simulate_charge_deposition(
            flux=flux,
            exposure=detector.time_step,
            x_lim=detector.geometry.horz_dimension,
            y_lim=detector.geometry.vert_dimension,
            z_lim=detector.geometry.total_thickness,
            step_size=step_size,
            energy_mean=energy_mean,
            energy_spread=energy_spread,
            energy_spectrum=energy_spectrum,
            energy_spectrum_sampling=energy_spectrum_sampling,
            ehpair_creation=ehpair_creation,
            material_density=material_density,
            particle_direction=particle_direction,
            stopping_power_curve=stopping_power_curve,
        )

    detector.charge.add_charge_dataframe(tracks_to_charge(tracks))


def tracks_to_charge(tracks: Sequence[float]) -> pd.DataFrame:
    """Convert tracks into a panda dataframe compatible with a detector charge object.

    Parameters
    ----------
    tracks : list
        with the following structure for each track :
        tracks[track_id], track_id = (vc, vx, vy, vz), vc[cluster_id] etc.
    """
    # tracks = np.array(tracks, dtype="object")

    tracks_arr = np.array(tracks, dtype="object")
    vc_arr = np.concatenate(tracks_arr[:, 0])
    vx_arr = np.concatenate(tracks_arr[:, 1])
    vy_arr = np.concatenate(tracks_arr[:, 2])
    vz_arr = np.concatenate(tracks_arr[:, 3])
    zeros = np.zeros(np.shape(vc_arr))

    return Charge.create_charges(
        particle_type="e",
        particles_per_cluster=vc_arr,
        init_energy=zeros,
        init_ver_position=vy_arr,
        init_hor_position=vx_arr,
        init_z_position=vz_arr,
        init_ver_velocity=zeros,
        init_hor_velocity=zeros,
        init_z_velocity=zeros,
    )


def _generate_random_energies(
    energy_mean: float,
    energy_spread: float,
    energy_spectrum: Union[str, Path, None],
    energy_spectrum_sampling: Literal["linear", "log"],
    n_p,
) -> np.ndarray:
    if energy_spectrum is None:
        # from normal distribution
        p_energies = np.random.normal(loc=energy_mean, scale=energy_spread, size=n_p)
    else:
        # from parsed spectrum using either logarithmic or linear sampling
        energy_spectrum_data = np.loadtxt(energy_spectrum, dtype="float", comments="#")
        n_samples = 100000
        if energy_spectrum_sampling == "log":
            log_energy_range = np.logspace(
                np.log10(np.min(energy_spectrum_data[:, 0])),
                np.log10(np.max(energy_spectrum_data[:, 0])),
                n_samples,
            )
            log_energy_pdf = np.interp(
                log_energy_range, energy_spectrum_data[:, 0], energy_spectrum_data[:, 1]
            )
            log_energy_pdf /= np.sum(log_energy_pdf)
            p_energies = np.random.choice(log_energy_range, size=n_p, p=log_energy_pdf)
        else:
            lin_energy_range = np.linspace(
                np.min(energy_spectrum_data[:, 0]),
                np.max(energy_spectrum_data[:, 0]),
                n_samples,
            )
            energy_pdf = np.interp(
                lin_energy_range, energy_spectrum_data[:, 0], energy_spectrum_data[:, 1]
            )
            energy_pdf /= np.sum(energy_pdf)
            p_energies = np.random.choice(lin_energy_range, n_p, p=energy_pdf)

    p_energies *= 1.0e6  # from MeV to eV

    return p_energies


def simulate_charge_deposition(
    flux: float,
    exposure: float,
    x_lim: float,
    y_lim: float,
    z_lim: float,
    step_size: float = 1.0,
    energy_mean: float = 1.0,
    energy_spread: float = 0.1,
    energy_spectrum: Union[str, Path, None] = None,
    energy_spectrum_sampling: Literal["linear", "log"] = "log",
    ehpair_creation: float = 3.65,
    material_density: float = 2.3290,
    particle_direction: Literal["isotropic", "orthogonal"] = "isotropic",
    stopping_power_curve: Union[str, Path, None] = None,
) -> list:
    """Simulate charge deposition of incident ionizing particles inside a detector.

    Parameters
    ----------
    flux : float
        the flux of incoming particles in particle/s
    exposure : float
        the detector exposure duration in s (how long it takes to take one readout)
    x_lim : float
        the maximum dimension of the sensor in the x direction
    y_lim : float
        the maximum dimension of the sensor in the y direction
    z_lim : float
        the maximum dimension of the sensor in the z direction (thickness)
    step_size : float
        the size of the considered unitary step in unit length along which energy is deposited
    energy_mean : float
        the mean energy of the incoming ionizing particles
    energy_spread : float
        the spread in energy of the incoming ionizing particles
    energy_spectrum : str
        the location of the file describing the energy spectrum of incident particles
        if no spectrum is provided energies are randomly drawn from a normal distribution
        with mean and spread defined above
        note that the energy spectrum is assumed to be a txt file with two columns [energy, flux]
        with the energy in MeV
    energy_spectrum_sampling : str. Default: 'log'
        "log" : the energy spectrum is sampled in log space
        "linear" : the energy spectrum is sampled in linear space
    ehpair_creation : float
        the energy required to generate a electron-hole pair in eV
        by default the Si value at room temperature is parsed i.e. 3.65 eV
    material_density : float
        the material density in g/cm3
        by default he Si value at room temperature is parsed i.e. 2.3290 g/cm3
    particle_direction : str
        "isotropic" : particles are coming from all directions (outside of the sensor)
        "orthogonal" : particles are coming from the top of the sensor (thickness = 0) and orthogonal to its surface
    stopping_power_curve : str
        the location of the file describing the total massive stopping power
        energetic loss per mass of material and per unit path length versus particle energy
        note that the the stopping power curve is assumed to be a csv file with two columns [energy, stopping power]
        energy in MeV, stopping power in MeV cm2/g
    """

    # determine the total number of ionizing particles to simulate based on flux and exposure duration
    n_p = np.int64(flux * exposure)

    if n_p <= 0:
        raise ValueError("The number of particles generated has to be greater than 0.")

    if x_lim <= 0 or y_lim <= 0 or z_lim <= 0:
        raise ValueError("Detector dimension is negative or 0.")

    # generate random energies for each particle
    p_energies: np.ndarray = _generate_random_energies(
        energy_mean=energy_mean,
        energy_spread=energy_spread,
        energy_spectrum=energy_spectrum,
        energy_spectrum_sampling=energy_spectrum_sampling,
        n_p=n_p,
    )

    # extract stopping power data convert to correct units and compute initial deposited energy
    if stopping_power_curve is None:
        raise ValueError("No stopping power curve has been parsed.")

    stopping_power_data = np.genfromtxt(
        stopping_power_curve, skip_header=1, delimiter=","
    )
    stopping_power_data[:, 0] *= 1.0e6  # from MeV to eV
    stopping_power_data[:, 1] *= (
        material_density * 1.0e2 * step_size
    )  # from MeV cm2/g to eV
    deposited_energies = np.interp(
        p_energies, stopping_power_data[:, 0], stopping_power_data[:, 1]
    )

    tracks = []
    # for each particle generate and store the energy deposition track
    for e, de in zip(p_energies, deposited_energies):
        # generate random particle coordinates
        x = np.random.random() * x_lim
        y = np.random.random() * y_lim

        # generate direction
        if particle_direction == "isotropic":
            alpha = 2.0 * np.pi * np.random.random()
            beta = 2.0 * np.pi * np.random.random()
        else:
            alpha = np.pi / 2.0
            beta = np.pi / 2.0
            # alpha = np.pi/2. + (np.random.randint(0,2))*np.pi
            # beta = np.pi/2. + (np.random.randint(0,2))*np.pi

        # compute equivalent step_size along each dimension
        dx = np.cos(alpha) * np.cos(beta) * step_size
        dy = np.cos(alpha) * np.sin(beta) * step_size
        dz = np.sin(alpha) * step_size

        # the particle is coming from outside the sensor: top or bottom, depending on travel direction (i.e. dz sign)
        z: float = 0.0 if dz >= 0 else z_lim

        # compute distances to the closest edge of the sensor along each dimension
        distance_x = (x_lim - x) if dx > 0 else x
        distance_y = (y_lim - y) if dy > 0 else y
        distance_z = (z_lim - z) if dz > 0 else z

        # compute the minimum number of steps required to reach one of the sensor edge
        distances = np.array([distance_x, distance_y, distance_z])
        deltas = np.abs(np.array([dx, dy, dz]))
        n_steps = int(np.min(np.divide(distances, deltas)))

        # check if incoming particle has enough energy to deposit all along the track or stops in the sensor
        if (e - n_steps * de) < 0:
            n_steps = int(e / de)
            # normally de should always be smaller than e
            # but it may happen that the particles would have stopped within a step_size
            # we assume that in that case the particle deposits its entire energy in one step
            if n_steps < 1:
                n_steps = 1
            vde = np.zeros(n_steps)
            for step in range(n_steps):
                vde[step] = np.interp(
                    e, stopping_power_data[:, 0], stopping_power_data[:, 1]
                )
                e -= de
        else:  # otherwise assumes the deposited energy is not a significant fraction of input energy,
            # the input energy remains the same all along track as os the deposited energy
            vde = np.ones(n_steps) * de

            # convert vde in amount of charge created
        vc = vde / ehpair_creation

        # generate vectors
        vx = x + np.arange(n_steps) * dx
        vy = y + np.arange(n_steps) * dy
        vz = z + np.arange(n_steps) * dz

        # gather results in a track
        track = (vc, vx, vy, vz)
        tracks.append(track)

    return tracks



# Copyright (c) 2023 Arthur Kadela and Joonas Viuho, Niels Bohr Institute, University of Copenhagen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Model for replicating the gain register in an EMCCD, including clock-induced-charge (CIC)."""
import numba
import numpy as np

from pyxel.detectors import CCD


def multiplication_register_cic(
    detector: CCD,
    total_gain: int,
    gain_elements: int,
    pcic_rate: float,
    scic_rate: float,
) -> None:
    """Calculate total gain of image with EMCCD multiplication register.

    Parameters
    ----------
    detector : CCD
    total_gain : int
    gain_elements : int
        Amount of single stage gain elements in the EMCCD register.
    pcic_rate : float
        Parallel CIC rate
    scic_rate : float
        Serial CIC rate
    """

    if total_gain < 0 or gain_elements < 0:
        raise ValueError("Wrong input parameter")

    # TODO: Set number of threads used by numba
    #       See: https://numba.readthedocs.io/en/stable/user/threading-layer.html#api-reference

    detector.pixel.array = multiplication_register_poisson(
        image_cube=detector.pixel.array,
        total_gain=total_gain,
        gain_elements=gain_elements,
        pcic_rate=pcic_rate,
        scic_rate=scic_rate,
    ).astype(float)


@numba.njit
def poisson_register(lam, new_image_cube_pix, gain_elements, scic_rate):
    """Calculate the total gain of a single pixel from EMCCD register elements.

    A single pixel is inputted and iterated through the total number of gain elements
    with the result being the resultant signal from the pixel going through the multiplication process.

    Parameters
    ----------
    lam : float
    new_image_cube_pix : int
    gain_elements : int
    scic_rate : float

    Returns
    -------
    int
    """

    new_image_cube_pix = new_image_cube_pix

    for _ in range(gain_elements):
        # Add possibility for a CIC event at each register stage
        new_image_cube_pix += np.random.poisson(scic_rate)

        # Each electron increase has chance for impact ionization, so one needs
        # to loop over all electrons at each gain stage.
        gain_electrons = 0
        for _ in range(np.floor(new_image_cube_pix)):
            gain_electrons += np.random.poisson(lam)
        new_image_cube_pix += gain_electrons

    return new_image_cube_pix


@numba.njit(parallel=True)
def multiplication_register_poisson(
    image_cube: np.ndarray,
    total_gain: int,
    gain_elements: int,
    pcic_rate: float,
    scic_rate: float,
) -> np.ndarray:
    """Calculate total gain of image from EMCCD register.

    Cycles through each pixel within the image provided.
    Returns a final image with signal added.

    Parameters
    ----------
    image_cube : np.ndarray
    total_gain : int
    gain_elements : int
        Amount of single stage gain elements in the EMCCD register.
    pcic_rate : float
        Parallel CIC rate.
    scic_rate : float
        Serial CIC rate.

    Returns
    -------
    np.ndarray
    """

    new_image_cube = np.zeros_like(image_cube, dtype=np.int32)

    # Generate and add pCIC to the frame
    pcic = np.random.poisson(pcic_rate, image_cube.shape)
    new_image_cube += pcic

    lam = total_gain ** (1 / gain_elements) - 1
    yshape, xshape = image_cube.shape

    for j in numba.prange(yshape):
        for i in numba.prange(xshape):
            new_image_cube[j, i] = poisson_register(
                lam=lam,
                new_image_cube_pix=image_cube[j, i],
                gain_elements=gain_elements,
                scic_rate=scic_rate,
            )

    return new_image_cube



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Utility functions."""

import logging


import numpy as np

from typing import Callable

# flake8: noqa
from .memory import get_size, memory_usage_details
from .examples import download_examples
from .timing import time_pipeline
from .add_model import create_model
from .randomize import set_random_seed
from .image import fit_into_array, load_cropped_and_aligned_image

__all__ = [
    "convert_to_int",
    "round_convert_to_int",
    "LogFilter",
    "load_cropped_and_aligned_image",
    "set_random_seed",
    "deprecated",
]


def round_convert_to_int(input_array: np.ndarray) -> np.ndarray:
    """Round list of floats in numpy array and convert to integers.

    Use on data before adding into DataFrame.

    :param input_array: numpy array object OR numpy array (float, int)
    :return:
    """
    array = input_array.astype(float)
    array = np.rint(array)
    array = array.astype(int)
    return array


def convert_to_int(input_array: np.ndarray) -> np.ndarray:
    """Convert numpy array to integer.

    Use on data after getting it from DataFrame.

    :param input_array: numpy array object OR numpy array (float, int)
    :return:
    """
    return input_array.astype(int)


class LogFilter(logging.Filter):
    """TBW."""

    def filter(self, record):
        """Filter or modify record."""
        if record.threadName == "MainThread" or record.threadName.endswith("-0"):
            return True

        return False


def deprecated(msg: str) -> Callable:
    """Deprecate a function.

    This decorator is based on (future ?) PEP 702.

    Examples
    --------
    >>> deprecated("Use 'new_function'")
    >>> def old_function(a, b):
    ...     return a + b
    ...

    >>> old_function.__deprecated__
    'Use new_function'
    >>> old_function(1, 2)
    DeprecationWarning: Use 'new_function'
      old_function(1, 2)
    """

    def _decorator(func: Callable) -> Callable:
        import warnings
        from functools import wraps

        @wraps(func)
        def _wrapper(*args, **kwargs):
            warnings.warn(msg, DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        # Add a new attribute '__deprecated__' to retrieve the error message (e.g. in 'ModelFunction')
        _wrapper.__deprecated__ = msg  # type: ignore
        return _wrapper

    return _decorator



#   Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#  #
#   This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#   is part of this Pyxel package. No part of the package, including
#   this file, may be copied, modified, propagated, or distributed except according to
#   the terms contained in the file ‘LICENCE.txt’.

import pytest
import xarray as xr
from datatree import DataTree

from pyxel.detectors import (
    CCD,
    CCDGeometry,
    Characteristics,
    Environment,
    ReadoutProperties,
)
from pyxel.models.data_processing import statistics


@pytest.fixture
def ccd_10x10() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )
    detector._readout_properties = ReadoutProperties(times=[1.0])
    return detector


def test_statistics(ccd_10x10: CCD):
    """Test input parameters for function 'statistics'."""
    detector = ccd_10x10
    statistics(detector=detector)

    data = detector.data
    assert isinstance(data, DataTree)

    for name in ["pixel", "photon", "signal", "image"]:
        data_statistics = data[f"/statistics/{name}"]
        assert isinstance(data_statistics, DataTree)

        assert "time" in data_statistics.coords
        assert list(data_statistics.coords["time"]) == [0.0]

        dataset = data_statistics.to_dataset()
        assert isinstance(dataset, xr.Dataset)

        assert "time" in dataset.coords
        assert list(dataset.coords["time"]) == [0.0]

    # Old tests
    # dataset = detector.processed_data.data
    # assert "time" in dataset.coords
    # assert list(dataset.coords["time"].values) == [0.0]



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


from typing import Literal

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from pyxel.detectors import CCD, CCDGeometry, Characteristics, Environment
from pyxel.models.readout_electronics import simple_adc
from pyxel.models.readout_electronics.simple_adc import _get_dtype, apply_simple_adc


@pytest.fixture
def ccd_3x3() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=3,
            col=3,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(
            adc_bit_resolution=16, adc_voltage_range=(0.0, 10.0)
        ),
    )
    return detector


@pytest.mark.parametrize("data_type", ["uint16", "uint32", "uint64", "uint"])
def test_simple_adc(
    ccd_3x3: CCD, data_type: Literal["uint16", "uint32", "uint64", "uint"]
):
    """Test model 'simple_adc'."""
    simple_adc(detector=ccd_3x3, data_type=data_type)


@pytest.mark.parametrize(
    "data_type, exp_exc, exp_error",
    [
        ("float", TypeError, "Expecting a signed/unsigned integer"),
        ("float32", TypeError, "Expecting a signed/unsigned integer"),
        ("float64", TypeError, "Expecting a signed/unsigned integer"),
        ("uint8", TypeError, "Expected type of Image array is uint64"),
        ("int32", TypeError, "Expected type of Image array is uint64"),
        ("int64", TypeError, "Expected type of Image array is uint64"),
        ("int", TypeError, "Expected type of Image array is uint64"),
    ],
)
def test_simple_adc_wrong_data_type(
    ccd_3x3: CCD,
    data_type: str,
    exp_exc,
    exp_error,
):
    """Test model 'simple_adc' with wrong 'data_type'."""
    with pytest.raises(exp_exc, match=exp_error):
        simple_adc(
            detector=ccd_3x3,
            data_type=data_type,
        )


@pytest.mark.parametrize(
    "signal, bit_resolution, voltage_range, dtype, exp_output",
    [
        pytest.param(
            np.array([0.0, 0.4, 0.5, 0.6, 1.0, 7.0, 8.0, 14.0, 15.0]),
            4,
            (0.0, 15.0),
            np.uint8,
            np.array([0, 0, 0, 0, 1, 7, 8, 14, 15], dtype=np.uint8),
            id="4 bit - Range: [0.0, 15.0]",
        ),
        pytest.param(
            np.array([-0.1, 0.0, 3.0, 6.0, 6.1]),
            8,
            (0.0, 6.0),
            np.uint8,
            np.array([0, 0, 127, 255, 255], dtype=np.uint8),
            id="8 bit - Range: [0.0, 6.0]",
        ),
        pytest.param(
            np.array([0.9, 1.0, 4.0, 7.0, 7.1]),
            8,
            (1.0, 7.0),
            np.uint8,
            np.array([0, 0, 127, 255, 255], dtype=np.uint8),
            id="8 bit - Range: [1.0, 7.0]",
        ),
        pytest.param(
            np.array([-1.1, -1.0, 2.0, 5.0, 5.1]),
            8,
            (-1.0, 5.0),
            np.uint8,
            np.array([0, 0, 127, 255, 255], dtype=np.uint8),
            id="8 bit - Range: [-1.0, 5.0]",
        ),
        pytest.param(
            np.array([-0.1, 0.0, 6.0, 6.1]),
            16,
            (0.0, 6.0),
            np.uint16,
            np.array([0, 0, 65535, 65535], dtype=np.uint16),
            id="16 bit - Range: [0.0, 6.0]",
        ),
        pytest.param(
            np.array([-0.1, 0.0, 6.0, 6.1]),
            32,
            (0.0, 6.0),
            np.uint32,
            np.array([0, 0, 2**32 - 1, 2**32 - 1], dtype=np.uint32),
            id="32 bit - Range: [0.0, 6.0]",
        ),
    ],
)
def test_apply_simple_adc(signal, bit_resolution, voltage_range, dtype, exp_output):
    """Test function 'apply_simple_adc'."""
    voltage_min, voltage_max = voltage_range
    output = apply_simple_adc(
        signal=signal,
        bit_resolution=bit_resolution,
        voltage_min=voltage_min,
        voltage_max=voltage_max,
        dtype=dtype,
    )

    assert_array_equal(output, exp_output)

    # TODO: Use param 'strict' in 'assert_array_equal' (available in Numpy 1.24)
    assert output.dtype == exp_output.dtype


@pytest.mark.parametrize(
    "bit_resolution, exp_dtype",
    [
        (1, np.dtype(np.uint8)),
        (8, np.dtype(np.uint8)),
        (9, np.dtype(np.uint16)),
        (16, np.dtype(np.uint16)),
        (17, np.dtype(np.uint32)),
        (32, np.dtype(np.uint32)),
        (33, np.dtype(np.uint64)),
        (64, np.dtype(np.uint64)),
    ],
)
def test_get_dtype(bit_resolution, exp_dtype):
    """Test function '_get_dtype'."""
    result = _get_dtype(bit_resolution)
    assert result == exp_dtype
